"""Object representing boolean API option."""
#
# (C) Pywikibot team, 2015-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from collections.abc import MutableMapping

import pywikibot
from pywikibot.backports import Set
from pywikibot.tools import deprecate_arg


__all__ = ['OptionSet']


class OptionSet(MutableMapping):

    """A class to store a set of options which can be either enabled or not.

    If it is instantiated with the associated site, module and parameter
    it will only allow valid names as options. If instantiated 'lazy
    loaded' it won't checks if the names are valid until the site has
    been set (which isn't required, but recommended). The site can only
    be set once if it's not None and after setting it, any site (even
    None) will fail.
    """

    @deprecate_arg('dict', 'data')  # since 9.0
    def __init__(self,
                 site: pywikibot.site.APISite | None = None,
                 module: str | None = None,
                 param: str | None = None,
                 data: dict | None = None) -> None:
        """Initializer.

        If a site is given, the module and param must be given too.

        .. versionchanged:: 9.0
           *dict* parameter was renamed to *data*.

        :param site: The associated site
        :param module: The module name which is used by paraminfo.
            (Ignored when site is None)
        :param param: The parameter name inside the module. That
            parameter must have a 'type' entry. (Ignored when site is
            None)
        :param data: The initializing data dict which is used for
            :meth:`from_dict`
        """
        self._site_set = False
        self._enabled: Set[str] = set()
        self._disabled: Set[str] = set()
        self._set_site(site, module, param)
        if data:
            self.from_dict(data)

    def _set_site(self,
                  site: pywikibot.site.APISite | None,
                  module: str | None,
                  param: str | None,
                  *,
                  clear_invalid: bool = False) -> None:
        """Set the site and valid names.

        As soon as the site has been not None, any subsequent calls will
        fail, unless there had been invalid names and a KeyError was
        thrown.

        :param site: The associated site
        :param module: The module name which is used by paraminfo.
        :param param: The parameter name inside the module. That
            parameter must have a 'type' entry.
        :param clear_invalid: Instead of throwing a KeyError, invalid
            names are silently removed from the options (disabled by
            default).
        """
        if self._site_set:
            raise TypeError('The site cannot be set multiple times.')
        # If the entries written to this are valid, it will never be
        # overwritten
        self._valid_enable = set()
        self._valid_disable = set()
        if site is None:
            return

        for type_value in site._paraminfo.parameter(module, param)['type']:
            if type_value[0] == '!':
                self._valid_disable.add(type_value[1:])
            else:
                self._valid_enable.add(type_value)
        if clear_invalid:
            self._enabled &= self._valid_enable
            self._disabled &= self._valid_disable
        else:
            invalid_names = ((self._enabled - self._valid_enable)
                             | (self._disabled - self._valid_disable))
            if invalid_names:
                raise KeyError('OptionSet already contains invalid name(s) '
                               '"{}"'.format('", "'.join(invalid_names)))
        self._site_set = True

    def from_dict(self, dictionary: dict[str, bool | None]) -> None:
        """Load options from the dict.

        The options are not cleared before. If changes have been made
        previously, but only the dict values should be applied it needs
        to be cleared first.

        :param dictionary: a dictionary containing for each entry either
            the value False, True or None. The names must be valid
            depending on whether they enable or disable the option. All
            names with the value None can be in either of the list.
        """
        enabled = set()
        disabled = set()
        removed = set()
        for name, value in dictionary.items():
            if value is True:
                enabled.add(name)
            elif value is False:
                disabled.add(name)
            elif value is None:
                removed.add(name)
            else:
                raise ValueError(f'Dict contains invalid value "{value}"')
        invalid_names = (
            (enabled - self._valid_enable) | (disabled - self._valid_disable)
            | (removed - self._valid_enable - self._valid_disable)
        )
        if invalid_names and self._site_set:
            raise ValueError('Dict contains invalid name(s) "{}"'.format(
                '", "'.join(invalid_names)))
        self._enabled = enabled | (self._enabled - disabled - removed)
        self._disabled = disabled | (self._disabled - enabled - removed)

    def __setitem__(self, name, value) -> None:
        """Set option to enabled, disabled or neither."""
        if value is True:
            if self._site_set and name not in self._valid_enable:
                raise KeyError(f'Invalid name "{name}"')
            self._enabled.add(name)
            self._disabled.discard(name)
        elif value is False:
            if self._site_set and name not in self._valid_disable:
                raise KeyError(f'Invalid name "{name}"')
            self._disabled.add(name)
            self._enabled.discard(name)
        elif value is None:
            if self._site_set and (name not in self._valid_enable
                                   or name not in self._valid_disable):
                raise KeyError(f'Invalid name "{name}"')
            self._enabled.discard(name)
            self._disabled.discard(name)
        else:
            raise ValueError(f'Invalid value "{value}"')

    def __getitem__(self, name) -> bool | None:
        """Return whether the option is enabled.

        :return: If the name has been set it returns whether it is
            enabled. Otherwise it returns None. If the site has been set
            it raises a KeyError if the name is invalid. Otherwise it
            might return a value even though the name might be invalid.
        """
        if name in self._enabled:
            return True
        if name in self._disabled:
            return False
        if (self._site_set or name in self._valid_enable
                or name in self._valid_disable):
            return None
        raise KeyError(f'Invalid name "{name}"')

    def __delitem__(self, name) -> None:
        """Remove the item by setting it to None."""
        self[name] = None

    def __iter__(self):
        """Iterate over each enabled and disabled option."""
        yield from self._enabled
        yield from self._disabled

    def api_iter(self):
        """Iterate over each option as they appear in the URL."""
        yield from self._enabled
        for disabled in self._disabled:
            yield f'!{disabled}'

    def __len__(self) -> int:
        """Return the number of enabled and disabled options."""
        return len(self._enabled) + len(self._disabled)



"""User output/logging functions.

Six output functions are defined. Each requires a ``msg`` argument
All of these functions generate a message to the log file if
logging is enabled (`-log` or `-debug` command line arguments).

The functions :func:`info` (alias :func:`output`), :func:`stdout`,
:func:`warning` and :func:`error` all display a message to the user
through the logger object; the only difference is the priority level,
which can be used by the application layer to alter the display. The
:func:`stdout` function should be used only for data that is the
"result" of a script, as opposed to information messages to the user.

The function :func:`log` by default does not display a message to the
user, but this can be altered by using the `-verbose` command line
option.

The function :func:`debug` only logs its messages, they are never
displayed on the user console. :func:`debug()` takes a required second
argument, which is a string indicating the debugging layer.

.. seealso::
   - :pyhow:`Logging HOWTO<logging>`
   - :python:`Logging Cookbook<howto/logging-cookbook.html>`
"""
#
# (C) Pywikibot team, 2010-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations


__all__ = (
    'CRITICAL', 'DEBUG', 'ERROR', 'INFO', 'WARNING', 'STDOUT', 'VERBOSE',
    'INPUT',
    'add_init_routine',
    'logoutput', 'info', 'output', 'stdout', 'warning', 'error', 'log',
    'critical', 'debug', 'exception',
)

import logging
import os
import sys

# logging levels
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING
from typing import Any

from pywikibot.backports import Callable


STDOUT = 16  #:
VERBOSE = 18  #:
INPUT = 25  #:
"""Three additional logging levels which are implemented beside
:const:`CRITICAL`, :const:`DEBUG`, :const:`ERROR`, :const:`INFO` and
:const:`WARNING`.

.. seealso:: :python:`Python Logging Levels<logging.html#logging-levels>`
"""

_init_routines: list[Callable[[], Any]] = []
_inited_routines: set[Callable[[], Any]] = set()


def add_init_routine(routine: Callable[[], Any]) -> None:
    """Add a routine to be run as soon as possible."""
    _init_routines.append(routine)


def _init() -> None:
    """Init any routines which have not already been called."""
    for init_routine in _init_routines:
        found = init_routine in _inited_routines  # prevent infinite loop
        _inited_routines.add(init_routine)
        if not found:
            init_routine()

    # Clear the list of routines to be inited
    _init_routines[:] = []  # the global variable is used with slice operator


def logoutput(msg: Any,
              *args: Any,
              level: int = INFO,
              **kwargs: Any) -> None:
    """Format output and send to the logging module.

    Dispatch helper function used by all the user-output convenience
    functions. It can be used to implement your own high-level output
    function with a different logging level.

    *msg* can contain special sequences to create colored output. These
    consist of the color name in angle bracket, e. g. <<lightpurple>>.
    <<default>> resets the color.

    *args* are the arguments which are merged into *msg* using the
    string formatting operator. It is passed unchanged to the logger.

    .. attention:: Only old ``%``-formatting is supported for *args* by
       the :pylib:`logging` module.

    Keyword arguments other than those listed below are also passed
    unchanged to the logger; so far, the only argument that is useful is
    ``exc_info=True``, which causes the log message to include an
    exception traceback.

    .. versionchanged:: 7.2
       Positional arguments for *decoder* and *newline* are deprecated;
       keyword arguments should be used.
    .. versionchanged:: 10.0
       *args* parameter can now given as formatting arguments directly
       to the logger.

    :param msg: The message to be printed.
    :param args: Passed as string formatter options to the logger.
    :param level: The logging level; supported by :func:`logoutput` only.
    :keyword bool newline: If newline is True (default), a line feed
        will be added after printing the msg.
    :keyword str layer: Suffix of the logger name separated by dot. By
        default no suffix is used.
    :keyword str decoder: If msg is bytes, this decoder is used to
        decode. Default is 'utf-8', fallback is 'iso8859-1'
    :param kwargs: For the other keyword arguments refer
        :python:`Logger.debug()<library/logging.html#logging.Logger.debug>`
    """
    # invoke any init routines
    if _init_routines:
        _init()

    # frame 0 is logoutput() in this module,
    # frame 1 is the convenience function (info(), etc.)
    # frame 2 is whatever called the convenience function
    frame = sys._getframe(2)
    module = os.path.basename(frame.f_code.co_filename)
    newline = kwargs.pop('newline', True)
    context = {'caller_name': frame.f_code.co_name,
               'caller_file': module,
               'caller_line': frame.f_lineno,
               'newline': ('\n' if newline else '')}
    context.update(kwargs.pop('extra', {}))

    decoder = kwargs.pop('decoder', 'utf-8')
    if isinstance(msg, bytes):
        try:
            msg = msg.decode(decoder)
        except UnicodeDecodeError:
            msg = msg.decode('iso8859-1')

    layer = kwargs.pop('layer', '')
    logger = logging.getLogger(('pywiki.' + layer).strip('.'))
    logger.log(level, msg, *args, extra=context, **kwargs)


def info(msg: Any = '', *args: Any, **kwargs: Any) -> None:
    """Output a message to the user with level :const:`INFO`.

    ``msg`` will be sent to stderr via :mod:`pywikibot.userinterfaces`.
    It may be omitted and a newline is printed in that case.
    The arguments are interpreted as for :func:`logoutput`. *args* can
    be uses as formatting arguments for all logging functions of this
    module. But this works for old ``%``-formatting only:

    >>> info('Hello %s', 'World')  # doctest: +SKIP
    Hello World
    >>> info('Pywikibot %(version)d', {'version': 10})  # doctest: +SKIP
    Pywikibot 10

    .. versionadded:: 7.2
       was renamed from :func:`output`. Positional arguments for
       *decoder* and *newline* are deprecated; keyword arguments should
       be used. Keyword parameter *layer* was added.
    .. versionchanged:: 10.0
       *args* parameter can now given as formatting arguments directly
       to the logger.

    .. seealso::
       :python:`Logger.info()<library/logging.html#logging.Logger.info>`
    """
    logoutput(msg, *args, **kwargs)


output = info
"""Synonym for :func:`info` for backward compatibility. The arguments are
interpreted as for :func:`logoutput`.

.. versionchanged:: 7.2
   was renamed to :func:`info`; `text` was renamed to `msg`; `msg`
   paramerer may be omitted; only keyword arguments are allowed except
   for `msg`. Keyword parameter *layer* was added.
.. seealso::
   :python:`Logger.info()<library/logging.html#logging.Logger.info>`
"""


def stdout(msg: Any = '', *args: Any, **kwargs: Any) -> None:
    """Output script results to the user with level :const:`STDOUT`.

    ``msg`` will be sent to standard output (stdout) via
    :mod:`pywikibot.userinterfaces`, so that it can be piped to another
    process. All other functions will send to stderr.
    `msg` may be omitted and a newline is printed in that case.

    The arguments are interpreted as for :func:`logoutput`.

    .. versionchanged:: 7.2
       `text` was renamed to `msg`; `msg` parameter may be omitted;
       only keyword arguments are allowed except for `msg`. Keyword
       parameter *layer* was added.
    .. versionchanged:: 10.0
       *args* parameter can now given as formatting arguments directly
       to the logger.
    .. seealso::
       - :python:`Logger.log()<library/logging.html#logging.Logger.log>`
       - :wiki:`Pipeline (Unix)`
    """
    logoutput(msg, *args, level=STDOUT, **kwargs)


def warning(msg: Any, *args: Any, **kwargs: Any) -> None:
    """Output a warning message to the user with level :const:`WARNING`.

    ``msg`` will be sent to stderr via :mod:`pywikibot.userinterfaces`.
    The arguments are interpreted as for :func:`logoutput`.

    .. versionchanged:: 7.2
       `text` was renamed to `msg`; only keyword arguments are allowed
       except for `msg`. Keyword parameter *layer* was added.
    .. versionchanged:: 10.0
       *args* parameter can now given as formatting arguments directly
       to the logger.
    .. seealso::
       :python:`Logger.warning()<library/logging.html#logging.Logger.warning>`
    """
    logoutput(msg, *args, level=WARNING, **kwargs)


def error(msg: Any, *args: Any, **kwargs: Any) -> None:
    """Output an error message to the user with level :const:`ERROR`.

    ``msg`` will be sent to stderr via :mod:`pywikibot.userinterfaces`.
    The arguments are interpreted as for :func:`logoutput`.

    .. versionchanged:: 7.2
       `text` was renamed to `msg`; only keyword arguments are allowed
       except for `msg`. Keyword parameter *layer* was added.
    .. versionchanged:: 10.0
       *args* parameter can now given as formatting arguments directly
       to the logger.
    .. seealso::
       :python:`Logger.error()<library/logging.html#logging.Logger.error>`
    """
    logoutput(msg, *args, level=ERROR, **kwargs)


def log(msg: Any, *args: Any, **kwargs: Any) -> None:
    """Output a record to the log file with level :const:`VERBOSE`.

    The arguments are interpreted as for :func:`logoutput`.

    .. versionchanged:: 7.2
       `text` was renamed to `msg`; only keyword arguments are allowed
       except for `msg`. Keyword parameter *layer* was added.
    .. versionchanged:: 10.0
       *args* parameter can now given as formatting arguments directly
       to the logger.
    .. seealso::
       :python:`Logger.log()<library/logging.html#logging.Logger.log>`
    """
    logoutput(msg, *args, level=VERBOSE, **kwargs)


def critical(msg: Any, *args: Any, **kwargs: Any) -> None:
    """Output a critical record to the user with level :const:`CRITICAL`.

    ``msg`` will be sent to stderr via :mod:`pywikibot.userinterfaces`.
    The arguments are interpreted as for :func:`logoutput`.

    .. versionchanged:: 7.2
       `text` was renamed to `msg`; only keyword arguments are allowed
       except for `msg`. Keyword parameter *layer* was added.
    .. versionchanged:: 10.0
       *args* parameter can now given as formatting arguments directly
       to the logger.
    .. seealso::
       :python:`Logger.critical()
       <library/logging.html#logging.Logger.critical>`
    """
    logoutput(msg, *args, level=CRITICAL, **kwargs)


def debug(msg: Any, *args: Any, **kwargs: Any) -> None:
    """Output a debug record to the log file with level :const:`DEBUG`.

    The arguments are interpreted as for :func:`logoutput`.

    .. versionchanged:: 7.2
       `layer` parameter is optional; `text` was renamed to `msg`;
       only keyword arguments are allowed except for `msg`.
    .. versionchanged:: 10.0
       *args* parameter can now given as formatting arguments directly
       to the logger.
    .. seealso::
       :python:`Logger.debug()<library/logging.html#logging.Logger.debug>`
    """
    logoutput(msg, *args, level=DEBUG, **kwargs)


def exception(msg: Any = None, *args: Any,
              exc_info: bool = True, **kwargs: Any) -> None:
    """Output an error traceback to the user with level :const:`ERROR`.

    Use directly after an 'except' statement::

        ...
        except Exception:
            pywikibot.exception()
        ...

    or alternatively::

        ...
        except Exception as e:
            pywikibot.exception(e)
        ...

    With `exc_info=False` this function works like :func:`error` except
    that the `msg` parameter may be omitted.
    This function should only be called from an Exception handler.
    ``msg`` will be sent to stderr via :mod:`pywikibot.userinterfaces`.
    The arguments are interpreted as for :func:`logoutput`.

    .. versionchanged:: 7.2
       only keyword arguments are allowed except for `msg`;
       `exc_info` keyword is to be used instead of `tb`. Keyword
       parameter *layer* was added.
    .. versionchanged:: 7.3
       `exc_info` is True by default
    .. versionchanged:: 10.0
       *args* parameter can now given as formatting arguments directly
       to the logger.
    .. seealso::
       :python:`Logger.exception()
       <library/logging.html#logging.Logger.exception>`

    The arguments are interpreted as for :meth:`output`.
    """
    if msg is None:
        exc_type, value, _tb = sys.exc_info()
        msg = str(value)
        if exc_type is not None and not exc_info:
            msg += f' ({exc_type.__name__})'
    assert msg is not None
    error(msg, *args, exc_info=exc_info, **kwargs)



"""Classes for detecting a MediaWiki site."""
#
# (C) Pywikibot team, 2010-2024
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import json
import re
from contextlib import suppress
from html.parser import HTMLParser
from http import HTTPStatus
from urllib.parse import urljoin, urlparse

from requests import JSONDecodeError
from requests.exceptions import RequestException

import pywikibot
from pywikibot.backports import removesuffix
from pywikibot.comms.http import fetch
from pywikibot.exceptions import ClientError, ServerError
from pywikibot.tools import MediaWikiVersion


SERVER_DB_ERROR_MSG = \
    '<h1>Sorry! This site is experiencing technical difficulties.</h1>'

MIN_VERSION = MediaWikiVersion('1.31')


class MWSite:

    """Minimal wiki site class."""

    def __init__(self, fromurl, **kwargs) -> None:
        """Initializer.

        :raises pywikibot.exceptions.ServerError: a server error
            occurred while loading the site
        :raises Timeout: a timeout occurred while loading the site
        :raises RuntimeError: Version not found or version less than
            1.31
        """
        fromurl = removesuffix(fromurl, '$1')

        r = fetch(fromurl, **kwargs)
        check_response(r)

        if fromurl != r.url:
            pywikibot.log(f'{fromurl} redirected to {r.url}')
            fromurl = r.url

        self.fromurl = fromurl

        data = r.text

        wp = WikiHTMLPageParser(fromurl)
        wp.feed(data)

        self.version = wp.version
        self.server = wp.server
        self.scriptpath = wp.scriptpath
        self.articlepath = None

        if self.api:
            try:
                self._parse_site()
            except (ServerError, RequestException):
                raise
            except Exception as e:
                pywikibot.log(f'MW detection failed: {e!r}')

            if not self.version:
                self._fetch_old_version()

        if not self.api:
            raise RuntimeError(f'Unsupported url: {self.fromurl}')

        if not self.version or self.version < MIN_VERSION:
            raise RuntimeError(f'Unsupported version: {self.version}')

        if not self.articlepath:
            if not self.private_wiki:
                raise RuntimeError(
                    f'Unable to determine articlepath: {self.fromurl}')

            if self.api == self.fromurl:
                raise RuntimeError(
                    'Unable to determine articlepath because the wiki is '
                    'private. Use the Main Page URL instead of the API.')

            self.articlepath = self.fromurl.rsplit('/', 1)[0] + '/$1'

    def __repr__(self) -> str:
        return f'{type(self).__name__}("{self.fromurl}")'

    @property
    def langs(self):
        """Build interwikimap."""
        response = fetch(
            self.api
            + '?action=query&meta=siteinfo&siprop=interwikimap'
              '&sifilteriw=local&format=json')
        iw = response.json()

        error = iw.get('error')
        if error:
            raise RuntimeError(f"{error['code']} - {error['info']}")

        return [wiki for wiki in iw['query']['interwikimap']
                if 'language' in wiki]

    def _fetch_old_version(self) -> None:
        """Extract the version from API help with ?version enabled."""
        if self.version is None:
            try:
                r = fetch(self.api + '?version&format=json')
                try:
                    d = r.json()
                except JSONDecodeError:
                    # Fallback for old versions which didn't wrap help in json
                    d = {'error': {'*': r.text}}

                self.version = list(filter(
                    lambda x: x.startswith('MediaWiki'),
                    (line.strip()
                     for line in d['error']['*'].split('\n'))))[0].split()[1]
            except Exception:
                pass
            else:
                self.version = MediaWikiVersion(self.version)

    def _parse_site(self) -> None:
        """Parse siteinfo data."""
        response = fetch(self.api + '?action=query&meta=siteinfo&format=json')
        check_response(response)
        # remove preleading newlines and Byte Order Mark (BOM), see T128992
        content = response.text.strip().lstrip('\uFEFF')
        info = json.loads(content)
        self.private_wiki = ('error' in info
                             and info['error']['code'] == 'readapidenied')
        if self.private_wiki:
            # user config is not loaded because PYWIKIBOT_NO_USER_CONFIG
            # is set to '2' by generate_family_file.py.
            # Prepare a temporary config for login.
            username = pywikibot.input(
                'Private wiki detected. Login is required.\n'
                'Please enter your username?')
            # Setup a dummy family so that we can create a site object
            fam = pywikibot.family.AutoFamily('temporary_family',
                                              self.server + self.scriptpath)
            site = pywikibot.Site(fam.code, fam, username)
            site.version = lambda: str(self.version)
            # Now the site object is able to login
            info = site.siteinfo
        else:
            info = info['query']['general']

        self.version = MediaWikiVersion.from_generator(info['generator'])
        if self.version < MIN_VERSION:
            return

        self.server = urljoin(self.fromurl, info['server'])
        for item in ['scriptpath', 'articlepath', 'lang']:
            setattr(self, item, info[item])

    def __eq__(self, other):
        """Return True if equal to other."""
        return (self.server + self.scriptpath
                == other.server + other.scriptpath)

    def __hash__(self):
        """Get hashable representation."""
        return hash(self.server + self.scriptpath)

    @property
    def api(self) -> str | None:
        """Get api URL."""
        if self.server is None or self.scriptpath is None:
            return None

        return self.server + self.scriptpath + '/api.php'

    @property
    def iwpath(self):
        """Get article path URL."""
        return self.server + self.articlepath


class WikiHTMLPageParser(HTMLParser):

    """Wiki HTML page parser."""

    def __init__(self, url) -> None:
        """Initializer."""
        super().__init__(convert_charrefs=True)
        self.url = urlparse(url)
        self.generator = None
        self.version = None
        self._parsed_url = None
        self.server = None
        self.scriptpath = None

    def set_version(self, value) -> None:
        """Set highest version."""
        if self.version and value < self.version:
            return

        self.version = value

    def set_api_url(self, url) -> None:
        """Set api_url."""
        url = url.split('.php', 1)[0]
        try:
            value, script_name = url.rsplit('/', 1)
        except ValueError:
            return

        if script_name not in ('api', 'load', 'opensearch_desc'):
            return

        if script_name == 'load':
            self.set_version(MediaWikiVersion('1.17.0'))
            if self._parsed_url:
                # A Resource Loader link is less reliable than other links.
                # Resource Loader can load resources from a different site.
                # e.g. http://kino.skripov.com/index.php/$1
                # loads resources from http://megawiki.net/
                return

        new_parsed_url = urlparse(value)
        if self._parsed_url:
            assert new_parsed_url.path == self._parsed_url.path

        if not new_parsed_url.scheme or not new_parsed_url.netloc:
            new_parsed_url = urlparse(
                f'{new_parsed_url.scheme or self.url.scheme}://'
                f'{new_parsed_url.netloc or self.url.netloc}'
                f'{new_parsed_url.path}'
            )
        elif self._parsed_url:
            # allow upgrades to https, but not downgrades
            if self._parsed_url.scheme == 'https' \
               and new_parsed_url.scheme != self._parsed_url.scheme:
                return

            # allow http://www.brickwiki.info/ vs http://brickwiki.info/
            if (new_parsed_url.netloc in self._parsed_url.netloc
                    or self._parsed_url.netloc in new_parsed_url.netloc):
                return

            assert new_parsed_url == self._parsed_url, \
                   f'{self._parsed_url} != {new_parsed_url}'

        self._parsed_url = new_parsed_url
        self.server = f'{self._parsed_url.scheme}://{self._parsed_url.netloc}'
        self.scriptpath = self._parsed_url.path

    def handle_starttag(self, tag, attrs) -> None:
        """Handle an opening tag."""
        attrs = dict(attrs)
        if tag == 'meta':
            if attrs.get('name') == 'generator':
                self.generator = attrs['content']
                with suppress(ValueError):
                    self.version = MediaWikiVersion.from_generator(
                        self.generator)
        elif tag == 'link' and 'rel' in attrs and 'href' in attrs:
            if attrs['rel'] in ('EditURI', 'stylesheet', 'search'):
                self.set_api_url(attrs['href'])
        elif tag == 'script' and 'src' in attrs:
            self.set_api_url(attrs['src'])


def check_response(response) -> None:
    """Raise ClientError or ServerError depending on http status.

    .. versionadded:: 3.0
    .. versionchanged:: 7.0
       Raise a generic :class:`exceptions.ServerError` if http status
       code is not IANA-registered but unofficial code
    .. versionchanged:: 8.1
       Raise a :class:`exceptions.ClientError` if status code is 4XX
    """
    for status_code, err_class, err_type in [
        (HTTPStatus.INTERNAL_SERVER_ERROR, ServerError, 'Server'),
        (HTTPStatus.BAD_REQUEST, ClientError, 'Client')
    ]:  # highest http status code first
        if response.status_code >= status_code:
            try:
                status = HTTPStatus(response.status_code)
            except ValueError as err:
                m = re.search(r'\d{3}', err.args[0], flags=re.ASCII)
                if not m:
                    raise
                msg = f'Generic {err_type} Error ({m.group()})'
            else:
                msg = f'({status}) {status.description}'

            raise err_class(msg)

    if response.status_code == HTTPStatus.OK \
       and SERVER_DB_ERROR_MSG in response.text:
        raise ServerError('Server cannot access the database')



"""Objects representing MediaWiki families."""
#
# (C) Pywikibot team, 2004-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import collections
import inspect
import logging
import string
import sys
import types
import urllib.parse as urlparse
import warnings
from importlib import import_module
from itertools import chain
from os.path import basename, dirname, splitext
from textwrap import fill
from typing import TYPE_CHECKING, NoReturn

import pywikibot
from pywikibot import config
from pywikibot.backports import DefaultDict, Mapping, Sequence, removesuffix
from pywikibot.data import wikistats
from pywikibot.exceptions import FamilyMaintenanceWarning, UnknownFamilyError
from pywikibot.tools import classproperty, deprecated


logger = logging.getLogger('pywiki.wiki.family')

if TYPE_CHECKING:
    CrossnamespaceType = DefaultDict[str, dict[str, list[int]]]

# Legal characters for Family.name and Family.langs keys
NAME_CHARACTERS = string.ascii_letters + string.digits
# nds_nl code alias requires "_"n
# dash must be the last char to be reused as regex
CODE_CHARACTERS = string.ascii_lowercase + string.digits + '_-'


class Family:

    """Parent singleton class for all wiki families.

    Families are immutable and initializer is unsupported. Any class
    modification should go to :meth:`__post_init__` class method.

    .. versionchanged:: 3.0
       the family class is immutable. Having an ``__init__`` initializer
       method a ``NotImplementedWarning`` will be given.
    .. versionchanged:: 8.0
       ``alphabetic``, ``alphabetic_revised`` and ``fyinterwiki``
       attributes where removed.
    .. versionchanged:: 8.2
       :attr:`obsolete` setter was removed.
    .. versionchanged:: 8.3
       Having an initializer method a ``FutureWarning`` will be given.
    .. versionchanged:: 9.0
       raises RuntimeError if an initializer method was found;
       :meth:`__post_init__` classmethod should be used instead.
    """

    def __new__(cls):
        """Allocator."""
        # any Family class defined in this file are abstract
        if cls in globals().values():
            raise TypeError(f'Abstract Family class {cls.__name__} cannot be'
                            ' instantiated;  subclass it instead')

        # Override classproperty
        cls.instance = super().__new__(cls)
        cls.__new__ = lambda cls: cls.instance  # shortcut

        # don't use hasattr() here. consider only the class itself
        if '__init__' in cls.__dict__:
            raise RuntimeError(fill(
                f'Family class {cls.__module__}.{cls.__name__} cannot be'
                ' instantiated; use __post_init__() classmethod to modify'
                ' your family class. Refer the documentation.', width=66))

        if '__post_init__' not in cls.__dict__:
            pass
        elif inspect.ismethod(cls.__post_init__):  # classmethod check
            cls.__post_init__()
        else:
            raise RuntimeError(fill(
                f'__post_init__() method of {cls.__module__}.{cls.__name__}'
                ' class or its superclass must be a classmethod. Please  check'
                ' your family file.', width=66))

        return cls.instance

    @classproperty
    def instance(cls):
        """Get the singleton instance.

        This is a placeholder to invoke allocator before it's allocated.
        Allocator will override this classproperty.
        """
        return cls()

    #: The family name
    name: str | None = None

    #: Not open for edits; stewards can still edit.
    closed_wikis: list[str] = []

    #: Completely removed sites.
    removed_wikis: list[str] = []

    code_aliases: dict[str, str] = {}
    """Code mappings which are only an alias, and there is no 'old' wiki.

    For all except 'nl_nds', subdomains do exist as a redirect, but that
    should not be relied upon.
    """

    langs: dict[str, str] = {}

    #: A list of category redirect template names in different languages.
    category_redirect_templates: dict[str, Sequence[str]] = {
        '_default': []
    }

    #: A list of disambiguation template names in different languages.
    disambiguationTemplates: dict[str, Sequence[str]] = {
        '_default': []
    }

    edit_restricted_templates: dict[str, tuple[str, ...]] = {}
    """A dict of tuples for different sites with names of templates that
    indicate an edit should be avoided."""

    archived_page_templates: dict[str, tuple[str, ...]] = {}
    """A dict of tuples for different sites with names of archive templates
    that indicate an edit of non-archive bots should be avoided."""

    #: A set of projects that share cross-project sessions.
    cross_projects: set[str] = set()

    #: A list with the name for cross-project cookies, default for
    #: wikimedia centralAuth extensions.
    cross_projects_cookies = ['centralauth_Session',
                              'centralauth_Token',
                              'centralauth_User']
    cross_projects_cookie_username = 'centralauth_User'

    #: A list with the name in the cross-language flag permissions.
    cross_allowed: list[str] = []

    disambcatname: dict[str, str] = {}
    """A dict with the name of the category containing disambiguation pages for
    the various languages. Only one category per language, and without the
    namespace, so add things like:

    'en': "Disambiguation"
    """

    interwiki_attop: list[str] = []
    """Attop is a list of languages that prefer to have the interwiki links at
    the top of the page."""

    interwiki_on_one_line: list[str] = []
    """on_one_line is a list of languages that want the interwiki links one-
    after-another on a single line."""

    #: String used as separator between interwiki links and the text.
    interwiki_text_separator = '\n\n'

    # Similar for category
    category_attop: list[str] = []
    """Attop is a list of categories that prefer to have the category links at
    the top of the page."""

    category_on_one_line: list[str] = []
    """on_one_line is a list of languages that want the category links one-
    after-another on a single line."""

    #: String used as separator between category links and the text
    category_text_separator = '\n\n'

    interwiki_putfirst: dict[str, str] = {}
    """Which languages have a special order for putting interlanguage links,
    and what order is it?

    If a language is not in interwiki_putfirst, alphabetical order on
    language code is used. For languages that are in interwiki_putfirst,
    interwiki_putfirst is checked first, and languages are put in the
    order given there. All other languages are put after those, in code-
    alphabetical order.
    """

    interwiki_forward: str | None = None
    """Some families, e.

    g. commons and meta, are not multilingual and forward interlanguage
    links to another family (wikipedia). These families can set this
    variable to the name of the target family.
    """

    #: Some languages belong to a group where the possibility is high
    #: that equivalent articles have identical titles among the group.
    language_groups = {
        # languages using the Arabic script
        'arab': [
            'ar', 'ary', 'arz', 'azb', 'ckb', 'fa', 'glk', 'ks', 'lrc',
            'mzn', 'ps', 'sd', 'ur',
            # languages using multiple scripts, including Arabic
            'ha', 'kk', 'ku', 'pnb', 'ug'
        ],
        # languages that use Chinese symbols
        'chinese': [
            'wuu', 'zh', 'zh-classical', 'zh-yue', 'gan', 'ii',
            # languages using multiple/mixed scripts, including Chinese
            'ja', 'za'
        ],
        # languages that use the Cyrillic alphabet
        'cyril': [
            'ab', 'av', 'ba', 'be', 'be-tarask', 'bg', 'bxr', 'ce', 'cu',
            'cv', 'kbd', 'koi', 'kv', 'ky', 'mk', 'lbe', 'mdf', 'mn', 'mo',
            'myv', 'mhr', 'mrj', 'os', 'ru', 'rue', 'sah', 'tg', 'tk',
            'udm', 'uk', 'xal',
            # languages using multiple scripts, including Cyrillic
            'ha', 'kk', 'sh', 'sr', 'tt'
        ],
        # languages that use a Greek script
        'grec': [
            'el', 'grc', 'pnt'
            # languages using multiple scripts, including Greek
        ],
        # languages that use the Latin alphabet
        'latin': [
            'aa', 'ace', 'af', 'ak', 'als', 'an', 'ang', 'ast', 'ay', 'bar',
            'bat-smg', 'bcl', 'bi', 'bm', 'br', 'bs', 'ca', 'cbk-zam', 'cdo',
            'ceb', 'ch', 'cho', 'chy', 'co', 'crh', 'cs', 'csb', 'cy', 'da',
            'de', 'diq', 'dsb', 'ee', 'eml', 'en', 'eo', 'es', 'et', 'eu',
            'ext', 'ff', 'fi', 'fiu-vro', 'fj', 'fo', 'fr', 'frp', 'frr',
            'fur', 'fy', 'ga', 'gag', 'gd', 'gl', 'gn', 'gv', 'hak', 'haw',
            'hif', 'ho', 'hr', 'hsb', 'ht', 'hu', 'hz', 'ia', 'id', 'ie', 'ig',
            'ik', 'ilo', 'io', 'is', 'it', 'jbo', 'jv', 'kaa', 'kab', 'kg',
            'ki', 'kj', 'kl', 'kr', 'ksh', 'kw', 'la', 'lad', 'lb', 'lg', 'li',
            'lij', 'lmo', 'ln', 'lt', 'ltg', 'lv', 'map-bms', 'mg', 'mh', 'mi',
            'ms', 'mt', 'mus', 'mwl', 'na', 'nah', 'nap', 'nds', 'nds-nl',
            'ng', 'nl', 'nn', 'no', 'nov', 'nrm', 'nv', 'ny', 'oc', 'om',
            'pag', 'pam', 'pap', 'pcd', 'pdc', 'pfl', 'pih', 'pl', 'pms', 'pt',
            'qu', 'rm', 'rn', 'ro', 'roa-rup', 'roa-tara', 'rw', 'sc', 'scn',
            'sco', 'se', 'sg', 'simple', 'sk', 'sl', 'sm', 'sn', 'so', 'sq',
            'srn', 'ss', 'st', 'stq', 'su', 'sv', 'sw', 'szl', 'tet', 'tl',
            'tn', 'to', 'tpi', 'tr', 'ts', 'tum', 'tw', 'ty', 'uz', 've',
            'vec', 'vi', 'vls', 'vo', 'wa', 'war', 'wo', 'xh', 'yo', 'zea',
            'zh-min-nan', 'zu',
            # languages using multiple scripts, including Latin
            'az', 'chr', 'ckb', 'ha', 'iu', 'kk', 'ku', 'rmy', 'sh', 'sr',
            'tt', 'ug', 'za'
        ],
        # Scandinavian languages
        'scand': [
            'da', 'fo', 'is', 'nb', 'nn', 'no', 'sv'
        ],
    }

    ldapDomain = ()
    """LDAP domain if your wiki uses LDAP authentication.

    .. seealso:: https://www.mediawiki.org/wiki/Extension:LDAPAuthentication2
    """

    crossnamespace: CrossnamespaceType = collections.defaultdict(dict)
    """Allows crossnamespace interwiki linking.

    Lists the possible crossnamespaces combinations; keys are
    originating namespace; values are dicts where keys are the
    originating langcode, or ``_default`` and  values are dicts where
    keys are the languages that can be linked to from the lang+ns, or
    ``_default``; values are a list of namespace numbers.

    **Examples:**

    Allowing linking *to* ``pt`` 102 namespace from any other lang 0
    namespace is:

    .. code-block:: Python

       crossnamespace[0] = {
           '_default': { 'pt': [102]}
       }

    While allowing linking *from* ``pt`` 102 namespace to any other
    lang 0 namespace is

    .. code-block:: Python

       crossnamespace[102] = {
           'pt': { '_default': [0]}
       }
    """

    shared_urlshortner_wiki: tuple[str, str] | None = None
    """Some wiki farms have UrlShortener extension enabled only on the main
    site.

    This value can specify this last one with
    ``(lang, family)`` tuple.
    """

    title_delimiter_and_aliases = ' _'
    """Titles usually are delimited by a space and the alias is replaced to
    this delimiter; e.g. "Main page" is the title with spaces as delimiters but
    "Main_page" also works. Other families may have different settings.

    .. note:: The first character is used as delimiter, the others are
       aliases.

    .. warning:: This attribute is used within ``re.sub()`` method. Use
       escape sequence if necessary

    .. versionadded:: 7.0
    """

    _families: dict[str, Family] = {}

    @classproperty
    @deprecated('site.has_extension("CategorySelect")', since='10.3.0')
    def categories_last(cls) -> list[str]:
        """Categories come after interwiki links for the given site codes.

        .. deprecated:: 10.3
           use :meth:`site.has_extension('CategorySelect')
           <pywikibot.site._apisite.APISite.has_extension>` instead
        """
        return []

    @staticmethod
    def load(fam: str | None = None):
        """Import the named family.

        :param fam: family name (if omitted, uses the configured
            default)
        :return: a Family instance configured for the named family.
        :raises pywikibot.exceptions.UnknownFamilyError: family not
            known
        """
        if fam is None:
            fam = config.family

        if fam in Family._families:
            return Family._families[fam]

        if fam not in config.family_files:
            raise UnknownFamilyError(f'Family {fam} does not exist')

        family_file = config.family_files[fam]

        if family_file.startswith(('http://', 'https://')):
            myfamily = AutoFamily(fam, family_file)
            Family._families[fam] = myfamily
            return Family._families[fam]

        try:
            # Ignore warnings due to dots in family names.
            # TODO: use more specific filter, so that family classes can use
            #     RuntimeWarning's while loading.
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', RuntimeWarning)
                sys.path.append(dirname(family_file))
                mod = import_module(splitext(basename(family_file))[0])
        except ImportError:
            raise UnknownFamilyError(f'Family {fam} does not exist')
        cls = mod.Family.instance
        if cls.name != fam:
            warnings.warn(f'Family name {cls.name} does not match family '
                          f'module name {fam}',
                          FamilyMaintenanceWarning,
                          stacklevel=2)
        # Family 'name' and the 'langs' codes must be ascii letters and digits,
        # and codes must be lower-case due to the Site loading algorithm;
        # codes can accept also underscore/dash.
        if not all(x in NAME_CHARACTERS for x in cls.name):
            warnings.warn(
                f'Name of family {cls.name} must be ASCII letters and digits'
                ' [a-zA-Z0-9]',
                FamilyMaintenanceWarning,
                stacklevel=2,
            )
        for code in cls.langs:
            if not all(x in CODE_CHARACTERS for x in code):
                warnings.warn(
                    f'Family {cls.name} code {code} must be ASCII lowercase'
                    ' letters and digits [a-z0-9] or underscore/dash [_-]',
                    FamilyMaintenanceWarning,
                    stacklevel=2,
                )
        Family._families[fam] = cls
        return cls

    def category_redirects(self, code, fallback: str = '_default'):
        """Return list of category redirect templates."""
        if not hasattr(self, '_catredirtemplates') \
           or code not in self._catredirtemplates:
            self._get_cr_templates(code, fallback)
        return self._catredirtemplates[code]

    def _get_cr_templates(self, code, fallback) -> None:
        """Build list of category redirect templates."""
        if not hasattr(self, '_catredirtemplates'):
            self._catredirtemplates = {}
        if code in self.category_redirect_templates:
            cr_template_tuple = self.category_redirect_templates[code]
        elif fallback and fallback in self.category_redirect_templates:
            cr_template_tuple = self.category_redirect_templates[fallback]
        else:
            self._catredirtemplates[code] = []
            return
        cr_set = set()
        site = pywikibot.Site(code, self)
        tpl_ns = site.namespaces.TEMPLATE
        for cr_template in cr_template_tuple:
            cr_page = pywikibot.Page(site, cr_template, ns=tpl_ns)
            # retrieve all redirects to primary template from API,
            # add any that are not already on the list
            for t in cr_page.backlinks(filter_redirects=True,
                                       namespaces=tpl_ns):
                newtitle = t.title(with_ns=False)
                if newtitle not in cr_template_tuple:
                    cr_set.add(newtitle)
        self._catredirtemplates[code] = list(cr_template_tuple) + list(cr_set)

    def get_edit_restricted_templates(self, code):
        """Return tuple of edit restricted templates.

        .. versionadded:: 3.0
        """
        return self.edit_restricted_templates.get(code, ())

    def get_archived_page_templates(self, code):
        """Return tuple of archived page templates.

        .. versionadded:: 3.0
        """
        return self.archived_page_templates.get(code, ())

    def disambig(self, code, fallback: str | None = '_default') -> list[str]:
        """Return list of disambiguation templates.

        :raises KeyError: unknown title for disambig template
        """
        if code in self.disambiguationTemplates:
            return self.disambiguationTemplates[code]

        if fallback:
            return self.disambiguationTemplates[fallback]

        raise KeyError(
            f'ERROR: title for disambig template in language {code} unknown')

    def protocol(self, code: str) -> str:
        """The protocol to use to connect to the site.

        May be overridden to return 'http'. Other protocols are not
        supported.

        .. versionchanged:: 8.2
           ``https`` is returned instead of ``http``.

        :param code: language code
        :return: protocol that this family uses
        """
        return 'https'

    def verify_SSL_certificate(self, code: str) -> bool:
        """Return whether a HTTPS certificate should be verified.

        .. versionadded:: 5.3
           renamed from ignore_certificate_error

        :param code: language code
        :return: flag to verify the SSL certificate;
                 set it to False to allow access if certificate has an error.
        """
        return True

    def hostname(self, code):
        """The hostname to use for standard http connections."""
        return self.langs[code]

    def ssl_hostname(self, code):
        """The hostname to use for SSL connections."""
        return self.hostname(code)

    def scriptpath(self, code: str) -> str:
        """The prefix used to locate scripts on this wiki.

        This is the value displayed when you enter {{SCRIPTPATH}} on a
        wiki page (often displayed at [[Help:Variables]] if the wiki has
        copied the master help page correctly).

        The default value is the one used on Wikimedia Foundation wikis,
        but needs to be overridden in the family file for any wiki that
        uses a different value.

        :param code: Site code
        :raises KeyError: code is not recognised
        :return: URL path without ending '/'
        """
        return '/w'

    def ssl_pathprefix(self, code) -> str:
        """The path prefix for secure HTTP access."""
        # Override this ONLY if the wiki family requires a path prefix
        return ''

    def _hostname(self, code, protocol=None):
        """Return the protocol and hostname."""
        if protocol is None:
            protocol = self.protocol(code)
        if protocol == 'https':
            host = self.ssl_hostname(code)
        else:
            host = self.hostname(code)
        return protocol, host

    def base_url(self, code: str, uri: str, protocol=None) -> str:
        """Prefix uri with port and hostname.

        :param code: The site code
        :param uri: The absolute path after the hostname
        :param protocol: The protocol which is used. If None it'll
            determine the protocol from the code.
        :return: The full URL ending with uri
        """
        protocol, host = self._hostname(code, protocol)
        if protocol == 'https':
            uri = self.ssl_pathprefix(code) + uri
        return urlparse.urljoin(f'{protocol}://{host}', uri)

    def path(self, code) -> str:
        """Return path to index.php."""
        return f'{self.scriptpath(code)}/index.php'

    def querypath(self, code) -> str:
        """Return path to query.php."""
        return f'{self.scriptpath(code)}/query.php'

    def apipath(self, code) -> str:
        """Return path to api.php."""
        return f'{self.scriptpath(code)}/api.php'

    def eventstreams_host(self, code) -> NoReturn:
        """Hostname for EventStreams.

        .. versionadded:: 3.0
        """
        raise NotImplementedError('This family does not support EventStreams')

    def eventstreams_path(self, code) -> NoReturn:
        """Return path for EventStreams.

        .. versionadded:: 3.0
        """
        raise NotImplementedError('This family does not support EventStreams')

    def get_address(self, code, title) -> str:
        """Return the path to title using index.php with redirects disabled."""
        return f'{self.path(code)}?{title=!s}&redirect=no'

    def interface(self, code: str) -> str:
        """Return interface to use for code."""
        if code in self.interwiki_removals:
            if code in self.codes:
                warnings.warn(f'Interwiki removal {code} is in {self} codes',
                              FamilyMaintenanceWarning, stacklevel=2)
            if code in self.closed_wikis:
                return 'ClosedSite'
            if code in self.removed_wikis:
                return 'RemovedSite'

        return config.site_interface

    def from_url(self, url: str) -> str | None:
        """Return whether this family matches the given url.

        It is first checking if a domain of this family is in the domain
        of the URL. If that is the case it's checking all codes and
        verifies that a path generated via :attr:`APISite.articlepath
        <pywikibot.site.APISite.articlepath>` and :attr:`Family.path`
        matches the path of the URL together with the hostname for that
        code.

        It is using :attr:`Family.domains` to first check if a domain
        applies and then iterates over :attr:`Family.codes` to actually
        determine which code applies.

        .. versionchanged:: 10.0
           *url* parameter does not have to contain a api/query/script
           path
        .. versionchanged:: 10.3
           accept a trailing slash in *url* after domain.

        :param url: the URL which may contain a ``$1``. If it's missing
            it is assumed to be at the end.
        :return: The language code of the URL. None if that URL is not
            from his family.
        :raises RuntimeError: When there are multiple languages in this
            family which would work with the given URL.
        """
        parsed = urlparse.urlparse(url)
        if parsed.scheme not in {'http', 'https', ''}:
            return None

        path = parsed.path
        if parsed.query:
            path += '?' + parsed.query

        # Discard $1 and everything after it
        path, *_ = path.partition('$1')

        for domain in self.domains:
            if domain in parsed.netloc:
                break
        else:
            return None

        matched_sites = set()
        for code in chain(self.codes,
                          getattr(self, 'test_codes', ()),
                          getattr(self, 'closed_wikis', ()),
                          ):
            if self._hostname(code)[1] == parsed.netloc:
                # Use the code and family instead of the url
                # This is only creating a Site instance if domain matches
                site = pywikibot.Site(code, self.name)
                pywikibot.log(f'Found candidate {site}')

                if path in ('', '/'):  # accept trailing slash
                    return site.code

                for iw_url in site._interwiki_urls():
                    iw_url, *_ = iw_url.partition('{}')
                    if path.startswith(iw_url):
                        matched_sites.add(site)
                        break

        if len(matched_sites) == 1:
            return matched_sites.pop().code

        if not matched_sites:
            return None

        raise RuntimeError(
            'Found multiple matches for URL "{}": {}'
            .format(url, ', '.join(str(s) for s in matched_sites)))

    @deprecated('config.maximum_GET_length', since='8.0.0')
    def maximum_GET_length(self, code):
        """Return the maximum URL length for GET instead of POST.

        .. deprecated:: 8.0
           Use :ref:`config.maximum_GET_length<Account Settings>` instead.
        """
        return config.maximum_GET_length

    def dbName(self, code) -> str:
        """Return the name of the MySQL database."""
        return f'{code}{self.name}'

    def encoding(self, code) -> str:
        """Return the encoding for a specific language wiki."""
        return 'utf-8'

    def encodings(self, code):
        """Return list of historical encodings for a specific language wiki."""
        return (self.encoding(code), )

    def __eq__(self, other):
        """Compare self with other.

        If other is not a Family() object, try to create one.
        """
        if not isinstance(other, Family):
            other = self.load(other)

        return self is other

    def __ne__(self, other):
        try:
            return not self.__eq__(other)
        except UnknownFamilyError:
            return False

    def __hash__(self):
        return hash(self.name)

    def __str__(self) -> str:
        assert isinstance(self.name, str)
        return self.name

    def __repr__(self) -> str:
        return f'Family("{self.name}")'

    def shared_image_repository(self, code):
        """Return the shared image repository, if any."""
        return (None, None)

    @deprecated(since='10.6.0')
    def isPublic(self, code) -> bool:
        """Check the wiki require logging in before viewing it.

        .. deprecated:: 10.6
        """
        return True

    def post_get_convert(self, site, getText):
        """Do a conversion on the retrieved text from the Wiki.

        For example a :wiki:`X-conversion in Esperanto
        <Esperanto_orthography#X-system>`.
        """
        return getText

    def pre_put_convert(self, site, putText):
        """Do a conversion on the text to insert on the Wiki.

        For example a :wiki:`X-conversion in Esperanto
        <Esperanto_orthography#X-system>`.
        """
        return putText

    @property
    def obsolete(self) -> types.MappingProxyType[str, str | None]:
        """Old codes that are not part of the family.

        Interwiki replacements override removals for the same code.

        :return: mapping of old codes to new codes (or None)
        """
        data = dict.fromkeys(self.interwiki_removals)
        data.update(self.code_aliases)
        return types.MappingProxyType(data)

    @classproperty
    def domains(cls) -> set[str]:
        """Get list of unique domain names included in this family.

        These domains may also exist in another family.
        """
        return set(cls.langs.values())

    @classproperty
    def codes(cls) -> set[str]:
        """Get list of codes used by this family."""
        return set(cls.langs.keys())

    @classproperty
    @deprecated('code_aliases', since='10.6.0')
    def interwiki_replacements(cls) -> Mapping[str, str]:
        """Return an interwiki code replacement mapping.

        Which language codes no longer exist and by which language code
        should they be replaced. If for example the language with code
        xx: now should get code yy:, add {'xx':'yy'} to
        :attr:`code_aliases`.

        .. deprecated:: 10.6
           Use :attr:`code_aliases` directly instead.
        .. versionchanged:: 8.2
           changed from dict to invariant mapping.
        """
        return types.MappingProxyType(cls.code_aliases)

    @classproperty
    def interwiki_removals(cls) -> frozenset[str]:
        """Return a list of interwiki codes to be removed from wiki pages.

        Codes that should be removed, usually because the site has been
        taken down.

        .. versionchanged:: 8.2
           changed from list to invariant frozenset.
        """
        return frozenset(cls.removed_wikis + cls.closed_wikis)


class SingleSiteFamily(Family):

    """Single site family."""

    def __new__(cls):
        """Initializer."""
        if not hasattr(cls, 'code'):
            cls.code = cls.name

        assert cls.domain

        cls.langs = {cls.code: cls.domain}

        return super().__new__(cls)

    @classproperty
    def domains(cls):
        """Return the full domain name of the site."""
        return (cls.domain, )

    def hostname(self, code):
        """Return the domain as the hostname."""
        return self.domain


class SubdomainFamily(Family):

    """Multi site wikis that are subdomains of the same top level domain."""

    def __new__(cls):
        """Initializer."""
        assert cls.domain
        return super().__new__(cls)

    @classproperty
    def langs(cls) -> dict[str, str]:
        """Property listing family languages."""
        codes = sorted(cls.codes)

        if hasattr(cls, 'test_codes'):
            codes += cls.test_codes

        codes += cls.closed_wikis

        # shortcut this classproperty
        cls.langs = {code: f'{code}.{cls.domain}' for code in codes}
        cls.langs.update({alias: f'{code}.{cls.domain}'
                          for alias, code in cls.code_aliases.items()})

        return cls.langs

    @classproperty
    def domains(cls):
        """Return the domain name of the sites in this family."""
        return [cls.domain]


class FandomFamily(Family):

    """Common features of Fandom families.

    .. versionadded:: 3.0
       renamed from WikiaFamily
    """

    @classproperty
    def langs(cls):
        """Property listing family languages."""
        codes = sorted(cls.codes)

        if hasattr(cls, 'code_aliases'):
            codes += cls.code_aliases

        return dict.fromkeys(codes, cls.domain)

    def scriptpath(self, code):
        """Return the script path for this family."""
        return '' if code == 'en' else ('/' + code)


class WikimediaFamily(Family):

    """Class for all wikimedia families.

    .. versionchanged:: 8.0
       :attr:`knows_codes` attribute was added.
    """

    multi_language_content_families = [
        'wikibooks',
        'wikinews',
        'wikipedia',
        'wikiquote',
        'wikisource',
        'wikiversity',
        'wikivoyage',
        'wiktionary',
    ]

    wikimedia_org_content_families = [
        'commons', 'incubator', 'species',
    ]

    wikimedia_org_meta_families = [
        'meta',
        'outreach',
        'strategy',
        'wikimediachapter',
        'wikimania',
    ]

    wikimedia_org_other_families = [
        'wikitech',
    ]

    other_content_families = [
        'lingualibre',
        'mediawiki',
        'wikidata',
        'wikifunctions',
    ]

    content_families = set(
        multi_language_content_families
        + wikimedia_org_content_families
        + other_content_families
    )

    wikimedia_org_families = set(
        wikimedia_org_content_families
        + wikimedia_org_meta_families
        + wikimedia_org_other_families
    )

    # CentralAuth cross available projects.
    cross_projects = set(
        multi_language_content_families
        + wikimedia_org_content_families
        + wikimedia_org_meta_families
        + other_content_families
    )

    # Known Wikimedia site codes
    known_codes = [
        'aa', 'ab', 'ace', 'ady', 'af', 'ak', 'als', 'alt', 'am', 'ami', 'an',
        'ang', 'ar', 'arc', 'ary', 'arz', 'as', 'ast', 'atj', 'av', 'avk',
        'awa', 'ay', 'az', 'azb', 'ba', 'ban', 'bar', 'bat-smg', 'bcl', 'be',
        'be-tarask', 'bg', 'bh', 'bi', 'bjn', 'blk', 'bm', 'bn', 'bo', 'bpy',
        'br', 'bs', 'bug', 'bxr', 'ca', 'cbk-zam', 'cdo', 'ce', 'ceb', 'ch',
        'cho', 'chr', 'chy', 'ckb', 'co', 'cr', 'crh', 'cs', 'csb', 'cu', 'cv',
        'cy', 'da', 'dag', 'de', 'din', 'diq', 'dk', 'dsb', 'dty', 'dv', 'dz',
        'ee', 'el', 'eml', 'en', 'eo', 'es', 'et', 'eu', 'ext', 'fa', 'ff',
        'fi', 'fiu-vro', 'fj', 'fo', 'fr', 'frp', 'frr', 'fur', 'fy', 'ga',
        'gag', 'gan', 'gcr', 'gd', 'gl', 'glk', 'gn', 'gom', 'gor', 'got',
        'gu', 'guw', 'gv', 'ha', 'hak', 'haw', 'he', 'hi', 'hif', 'ho', 'hr',
        'hsb', 'ht', 'hu', 'hy', 'hyw', 'hz', 'ia', 'id', 'ie', 'ig', 'ii',
        'ik', 'ilo', 'inh', 'io', 'is', 'it', 'iu', 'ja', 'jam', 'jbo', 'jv',
        'ka', 'kaa', 'kab', 'kbd', 'kbp', 'kcg', 'kg', 'ki', 'kj', 'kk', 'kl',
        'km', 'kn', 'ko', 'koi', 'kr', 'krc', 'ks', 'ksh', 'ku', 'kv', 'kw',
        'ky', 'la', 'lad', 'lb', 'lbe', 'lez', 'lfn', 'lg', 'li', 'lij', 'lld',
        'lmo', 'ln', 'lo', 'lrc', 'lt', 'ltg', 'lv', 'mad', 'mai', 'map-bms',
        'mdf', 'mg', 'mh', 'mhr', 'mi', 'min', 'mk', 'ml', 'mn', 'mni', 'mnw',
        'mo', 'mr', 'mrj', 'ms', 'mt', 'mus', 'mwl', 'my', 'myv', 'mzn', 'na',
        'nah', 'nan', 'nap', 'nb', 'nds', 'nds-nl', 'ne', 'new', 'ng', 'nia',
        'nl', 'nn', 'no', 'nov', 'nqo', 'nrm', 'nso', 'nv', 'ny', 'oc', 'olo',
        'om', 'or', 'os', 'pa', 'pag', 'pam', 'pap', 'pcd', 'pcm', 'pdc',
        'pfl', 'pi', 'pih', 'pl', 'pms', 'pnb', 'pnt', 'ps', 'pt', 'pwn', 'qu',
        'rm', 'rmy', 'rn', 'ro', 'roa-rup', 'roa-tara', 'ru', 'rue', 'rw',
        'sa', 'sah', 'sat', 'sc', 'scn', 'sco', 'sd', 'se', 'sg', 'sh', 'shi',
        'shn', 'si', 'simple', 'sk', 'skr', 'sl', 'sm', 'smn', 'sn', 'so',
        'sq', 'sr', 'srn', 'ss', 'st', 'stq', 'su', 'sv', 'sw', 'szl', 'szy',
        'ta', 'tay', 'tcy', 'te', 'tet', 'tg', 'th', 'ti', 'tk', 'tl', 'tn',
        'to', 'tpi', 'tr', 'trv', 'ts', 'tt', 'tum', 'tw', 'ty', 'tyv', 'udm',
        'ug', 'uk', 'ur', 'uz', 've', 'vec', 'vep', 'vi', 'vls', 'vo', 'wa',
        'war', 'wo', 'wuu', 'xal', 'xh', 'xmf', 'yi', 'yo', 'za', 'zea', 'zh',
        'zh-classical', 'zh-cn', 'zh-min-nan', 'zh-tw', 'zh-yue', 'zu',
    ]

    # Code mappings which are only an alias, and there is no 'old' wiki.
    # For all except 'nl_nds', subdomains do exist as a redirect, but that
    # should not be relied upon.
    code_aliases = {
        # Country aliases, see T87002
        'dk': 'da',
        'jp': 'ja',

        # Language aliases
        'gsw': 'als',  # T399411
        'lzh': 'zh-classical',  # T399697
        'nb': 'no',  # T86924
        'rup': 'roa-rup',  # T399693
        'sgs': 'bat-smg',  # T399438
        'vro': 'fiu-vro',  # T399444

        # Closed wiki redirection aliases
        'mo': 'ro',

        # Incomplete language code change, see T86915
        'minnan': 'zh-min-nan',
        'nan': 'zh-min-nan',

        'zh-tw': 'zh',
        'zh-cn': 'zh',

        # Miss-spelling
        'nds_nl': 'nds-nl',

        # Renamed, see T11823
        'be-x-old': 'be-tarask',
    }

    # WikimediaFamily uses Wikibase for the category name containing
    # disambiguation pages for the various languages. We need the
    # Wikibase code and item number:
    disambcatname = {'wikidata': 'Q1982926'}

    # UrlShortener extension is only usable on metawiki, and this wiki can
    # process links to all WM domains.
    shared_urlshortner_wiki = ('meta', 'meta')

    @classproperty
    def domain(cls):
        """Domain property."""
        if cls.name in (cls.multi_language_content_families
                        + cls.other_content_families):
            return cls.name + '.org'
        if cls.name in cls.wikimedia_org_families:
            return 'wikimedia.org'

        raise NotImplementedError(
            f"Family {cls.name} needs to define property 'domain'")

    def shared_image_repository(self, code):
        """Return Wikimedia Commons as the shared image repository."""
        return ('commons', 'commons')

    def eventstreams_host(self, code) -> str:
        """Return 'https://stream.wikimedia.org' as the stream hostname."""
        return 'https://stream.wikimedia.org'

    def eventstreams_path(self, code) -> str:
        """Return path for EventStreams."""
        return '/v2/stream'

    @property
    def languages_by_size(self) -> list[str]:
        """Language codes of the largest wikis.

        They should be roughly sorted by size.

        .. versionchanged:: 9.0
           Sorting order is retrieved via :mod:`wikistats` for each call.

        :raises NotImplementedError: Family is not member of
            :attr:`multi_language_content_families`
        """
        if self.name not in self.multi_language_content_families:
            raise NotImplementedError(
                f'languages_by_size is not implemented for {self.name} family')

        exceptions = {
            'wikiversity': ['beta']
        }

        ws = wikistats.WikiStats()
        table = ws.languages_by_size(self.name)
        assert type(self.obsolete).__name__ == 'mappingproxy', (
            f'obsolete attribute is of type {type(self.obsolete).__name__} but'
            ' mappingproxy was expected'
        )

        lbs = [
            code for code in table
            if not (code in self.obsolete
                    or code in exceptions.get(self.name, []))
        ]

        # add codes missing by wikistats
        missing = set(self.codes) - set(lbs)
        return lbs + list(missing)


class WikimediaOrgFamily(SingleSiteFamily, WikimediaFamily):

    """Single site family for sites hosted at ``*.wikimedia.org``."""

    @classproperty
    def domain(cls) -> str:
        """Return the parents domain with a subdomain prefix."""
        return f'{cls.name}.wikimedia.org'


class WikibaseFamily(Family):

    """A base class for a Wikibase Family.

    .. versionadded:: 8.2
    """

    def interface(self, code) -> str:
        """Return 'DataSite' for Wikibase family."""
        return 'DataSite'

    def entity_sources(self, code: str) -> dict[str, tuple[str, str]]:
        """Provide repository site information for entity types.

        The result must be structured as follows:

            {<entity type>: (<family code>, <family name>)}

        for example:

            {'property': ('test', 'wikidata')}

        If an empty dict is returned, all entity types are found in the
        current ``DataSite``.

        The result is used by :meth:`DataSite.get_repo_for_entity_type
        <pywikibot.site._datasite.DataSite.get_repo_for_entity_type>`
        """
        return {}


class DefaultWikibaseFamily(WikibaseFamily):

    """A base class for a Wikimedia Wikibase Family.

    This class holds defaults for :meth:`calendarmodel`,
    :meth:`default_globe` and :meth:`globes` to prevent code duplication.

    .. warning:: Possibly you have to adjust the repository site in
       :meth:`WikibaseFamily.entity_sources` to get the valid entity.

    .. versionadded:: 8.2
    """

    @property
    def interval_start_property(self) -> str:
        """Return the property for the start of an interval."""
        return 'P580'

    @property
    def interval_end_property(self) -> str:
        """Return the property for the end of an interval."""
        return 'P582'

    def calendarmodel(self, code) -> str:
        """Default calendar model for WbTime datatype."""
        return 'http://www.wikidata.org/entity/Q1985727'

    def default_globe(self, code) -> str:
        """Default globe for Coordinate datatype."""
        return 'earth'

    def globes(self, code):
        """Supported globes for Coordinate datatype."""
        return {
            'ariel': 'http://www.wikidata.org/entity/Q3343',
            'bennu': 'http://www.wikidata.org/entity/Q11558',
            'callisto': 'http://www.wikidata.org/entity/Q3134',
            'ceres': 'http://www.wikidata.org/entity/Q596',
            'deimos': 'http://www.wikidata.org/entity/Q7548',
            'dione': 'http://www.wikidata.org/entity/Q15040',
            'earth': 'http://www.wikidata.org/entity/Q2',
            'enceladus': 'http://www.wikidata.org/entity/Q3303',
            'eros': 'http://www.wikidata.org/entity/Q16711',
            'europa': 'http://www.wikidata.org/entity/Q3143',
            'ganymede': 'http://www.wikidata.org/entity/Q3169',
            'gaspra': 'http://www.wikidata.org/entity/Q158244',
            'hyperion': 'http://www.wikidata.org/entity/Q15037',
            'iapetus': 'http://www.wikidata.org/entity/Q17958',
            'io': 'http://www.wikidata.org/entity/Q3123',
            'jupiter': 'http://www.wikidata.org/entity/Q319',
            'lutetia': 'http://www.wikidata.org/entity/Q107556',
            'mars': 'http://www.wikidata.org/entity/Q111',
            'mercury': 'http://www.wikidata.org/entity/Q308',
            'mimas': 'http://www.wikidata.org/entity/Q15034',
            'miranda': 'http://www.wikidata.org/entity/Q3352',
            'moon': 'http://www.wikidata.org/entity/Q405',
            'oberon': 'http://www.wikidata.org/entity/Q3332',
            'phobos': 'http://www.wikidata.org/entity/Q7547',
            'phoebe': 'http://www.wikidata.org/entity/Q17975',
            'pluto': 'http://www.wikidata.org/entity/Q339',
            'rhea': 'http://www.wikidata.org/entity/Q15050',
            'ryugu': 'http://www.wikidata.org/entity/Q1385178',
            'steins': 'http://www.wikidata.org/entity/Q150249',
            'tethys': 'http://www.wikidata.org/entity/Q15047',
            'titan': 'http://www.wikidata.org/entity/Q2565',
            'titania': 'http://www.wikidata.org/entity/Q3322',
            'triton': 'http://www.wikidata.org/entity/Q3359',
            'umbriel': 'http://www.wikidata.org/entity/Q3338',
            'venus': 'http://www.wikidata.org/entity/Q313',
            'vesta': 'http://www.wikidata.org/entity/Q3030',
        }


def AutoFamily(name: str, url: str) -> SingleSiteFamily:
    """Family that automatically loads the site configuration.

    :param name: Name for the family
    :param url: API endpoint URL of the wiki
    :return: Generated family class
    """
    url = urlparse.urlparse(url)
    domain = url.netloc

    def protocol(self, code):
        """Return the protocol of the URL."""
        return self.url.scheme

    def scriptpath(self, code):
        """Extract the script path from the URL."""
        if self.url.path.endswith('/api.php'):
            return removesuffix(self.url.path, '/api.php')

        # AutoFamily refers to the variable set below, not the function
        # but the reference must be given here
        return super(AutoFamily, self).scriptpath(code)

    AutoFamily = type('AutoFamily', (SingleSiteFamily,), locals())
    return AutoFamily()



"""Special bot library containing BaseUnlinkBot.

Do not import classes directly from here but from specialbots.
"""
#
# (C) Pywikibot team, 2003-2024
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from pywikibot.bot import AutomaticTWSummaryBot, ExistingPageBot
from pywikibot.bot_choice import (
    AlwaysChoice,
    ChoiceException,
    InteractiveReplace,
    UnhandledAnswer,
)
from pywikibot.editor import TextEditor
from pywikibot.textlib import replace_links


class EditReplacementError(ChoiceException, UnhandledAnswer):

    """The text should be edited and replacement should be restarted."""

    def __init__(self) -> None:
        """Initializer."""
        super().__init__('edit', 'e', stop=True)


class InteractiveUnlink(InteractiveReplace):

    """An implementation which just allows unlinking."""

    def __init__(self, bot) -> None:
        """Create default settings."""
        super().__init__(old_link=bot.pageToUnlink,
                         new_link=False, default='u')
        self._always = AlwaysChoice(self, 'unlink all pages', 'a')
        self._always.always = bot.opt.always
        self.additional_choices = [
            AlwaysChoice(self, 'unlink all on page', 'p'),
            self._always, EditReplacementError()]
        self._bot = bot
        self.context = 100
        self.context_change = 100

    def handle_answer(self, choice):
        """Handle choice and store in bot's options."""
        answer = super().handle_answer(choice)
        self._bot.opt.always = self._always.always
        return answer


class BaseUnlinkBot(ExistingPageBot, AutomaticTWSummaryBot):

    """A basic bot unlinking a given link from the current page."""

    use_redirects = False

    def __init__(self, **kwargs) -> None:
        """Redirect all parameters and add namespace as an available option."""
        self.available_options.update({
            'namespaces': [],
            # Which namespaces should be processed?
            # default to [] which means all namespaces will be processed
        })
        super().__init__(**kwargs)

    def _create_callback(self):
        """Create a new callback instance for replace_links."""
        return InteractiveUnlink(self)

    def unlink(self, target_page) -> None:
        """Unlink all links linking to the target page."""
        text = self.current_page.text
        while True:
            unlink_callback = self._create_callback()
            try:
                text = replace_links(text, unlink_callback, target_page.site)
            except EditReplacementError:
                new_text = TextEditor().edit(
                    unlink_callback.current_text,
                    jumpIndex=unlink_callback.current_range[0])
                # if user didn't press Cancel
                text = new_text or unlink_callback.current_text
            else:
                break

        self.put_current(text)



"""Object representing a Wiki user."""
#
# (C) Pywikibot team, 2009-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import pywikibot
from pywikibot.backports import Generator
from pywikibot.exceptions import (
    APIError,
    AutoblockUserError,
    NoRenameTargetError,
    NotEmailableError,
    UserRightsError,
)
from pywikibot.page._links import Link
from pywikibot.page._page import Page
from pywikibot.page._revision import Revision
from pywikibot.tools import is_ip_address, is_ip_network


__all__ = ('User', )


class User(Page):

    """A class that represents a Wiki user.

    This class also represents the Wiki page User:<username>
    """

    def __init__(self, source, title: str = '') -> None:
        """Initializer for a User object.

        All parameters are the same as for Page() Initializer.
        """
        self._isAutoblock = True
        if title.startswith('#'):
            title = title[1:]
        elif ':#' in title:
            title = title.replace(':#', ':')
        else:
            self._isAutoblock = False
        super().__init__(source, title, ns=2)
        if self.namespace() != 2:
            raise ValueError(f"'{self.title()}' is not in the user namespace!")
        if self._isAutoblock:
            # This user is probably being queried for purpose of lifting
            # an autoblock.
            pywikibot.info(
                'This is an autoblock ID, you can only use to unblock it.')

    @property
    def username(self) -> str:
        """The username.

        Convenience method that returns the title of the page with
        namespace prefix omitted, which is the username.
        """
        if self._isAutoblock:
            return '#' + self.title(with_ns=False)
        return self.title(with_ns=False)

    def isRegistered(self, force: bool = False) -> bool:  # noqa: N802
        """Determine if the user is registered on the site.

        It is possible to have a page named User:xyz and not have
        a corresponding user with username xyz.

        The page does not need to exist for this method to return
        True.

        .. seealso:: :meth:`isAnonymous`

        :param force: if True, forces reloading the data from API
        """
        # T135828: the registration timestamp may be None but the key exists
        return (not self.isAnonymous()
                and 'registration' in self.getprops(force))

    def isAnonymous(self) -> bool:  # noqa: N802
        """Determine if the user is editing as an IP address.

        .. seealso::
           - :meth:`isRegistered`
           - :meth:`is_CIDR`
           - :func:`tools.is_ip_address`
        """
        return is_ip_address(self.username)

    def is_CIDR(self) -> bool:  # noqa: N802
        """Determine if the input refers to a range of IP addresses.

        .. versionadded:: 9.0
        .. seealso::
           - :meth:`isRegistered`
           - :meth:`isAnonymous`
           - :func:`tools.is_ip_network`
        """
        return is_ip_network(self.username)

    def getprops(self, force: bool = False) -> dict:
        """Return a properties about the user.

        .. versionchanged:: 9.0
           detect range blocks

        :param force: if True, forces reloading the data from API
        """
        if force and hasattr(self, '_userprops'):
            del self._userprops
        if not hasattr(self, '_userprops'):
            self._userprops = next(self.site.users([self.username]))
            if self.isAnonymous() or self.is_CIDR():
                r = next(self.site.blocks(iprange=self.username, total=1),
                         None)
                if r:
                    self._userprops['blockedby'] = r['by']
                    self._userprops['blockreason'] = r['reason']
        return self._userprops

    def registration(self,
                     force: bool = False) -> pywikibot.Timestamp | None:
        """Fetch registration date for this user.

        :param force: if True, forces reloading the data from API
        """
        if not self.isAnonymous():
            reg = self.getprops(force).get('registration')
            if reg:
                return pywikibot.Timestamp.fromISOformat(reg)
        return None

    def editCount(self, force: bool = False) -> int:  # noqa: N802
        """Return edit count for a registered user.

        Always returns 0 for 'anonymous' users.

        :param force: if True, forces reloading the data from API
        """
        return self.getprops(force).get('editcount', 0)

    def is_blocked(self, force: bool = False) -> bool:
        """Determine whether the user is currently blocked.

        .. versionchanged:: 7.0
           renamed from :meth:`isBlocked` method
        .. versionchanged:: 9.0
           can also detect range blocks.

        :param force: if True, forces reloading the data from API
        """
        return 'blockedby' in self.getprops(force)

    def is_locked(self, force: bool = False) -> bool:
        """Determine whether the user is currently locked globally.

        .. versionadded:: 7.0

        :param force: if True, forces reloading the data from API
        """
        return self.site.is_locked(self.username, force)

    def isEmailable(self, force: bool = False) -> bool:  # noqa: N802
        """Determine whether emails may be send to this user through MediaWiki.

        :param force: if True, forces reloading the data from API
        """
        return not self.isAnonymous() and 'emailable' in self.getprops(force)

    def groups(self, force: bool = False) -> list:
        """Return a list of groups to which this user belongs.

        The list of groups may be empty.

        :param force: if True, forces reloading the data from API
        :return: groups property
        """
        return self.getprops(force).get('groups', [])

    def gender(self, force: bool = False) -> str:
        """Return the gender of the user.

        :param force: if True, forces reloading the data from API
        :return: return 'male', 'female', or 'unknown'
        """
        if self.isAnonymous():
            return 'unknown'
        return self.getprops(force).get('gender', 'unknown')

    def rights(self, force: bool = False) -> list:
        """Return user rights.

        :param force: if True, forces reloading the data from API
        :return: return user rights
        """
        return self.getprops(force).get('rights', [])

    def getUserPage(self, subpage: str = '') -> Page:  # noqa: N802
        """Return a Page object relative to this user's main page.

        :param subpage: subpage part to be appended to the main page
            title (optional)
        :return: Page object of user page or user subpage
        """
        if self._isAutoblock:
            # This user is probably being queried for purpose of lifting
            # an autoblock, so has no user pages per se.
            raise AutoblockUserError(
                'This is an autoblock ID, you can only use to unblock it.')
        if subpage:
            subpage = '/' + subpage
        return Page(Link(self.title() + subpage, self.site))

    def getUserTalkPage(self, subpage: str = '') -> Page:  # noqa: N802
        """Return a Page object relative to this user's main talk page.

        :param subpage: subpage part to be appended to the main talk
            page title (optional)
        :return: Page object of user talk page or user talk subpage
        """
        if self._isAutoblock:
            # This user is probably being queried for purpose of lifting
            # an autoblock, so has no user talk pages per se.
            raise AutoblockUserError(
                'This is an autoblock ID, you can only use to unblock it.')
        if subpage:
            subpage = '/' + subpage
        return Page(Link(self.username + subpage,
                         self.site, default_namespace=3))

    def send_email(self, subject: str, text: str, ccme: bool = False) -> bool:
        """Send an email to this user via MediaWiki's email interface.

        :param subject: the subject header of the mail
        :param text: mail body
        :param ccme: if True, sends a copy of this email to the bot
        :raises NotEmailableError: the user of this User is not
            emailable
        :raises UserRightsError: logged in user does not have
            'sendemail' right
        :return: operation successful indicator
        """
        if not self.isEmailable():
            raise NotEmailableError(self)

        if not self.site.has_right('sendemail'):
            raise UserRightsError("You don't have permission to send mail")

        params = {
            'action': 'emailuser',
            'target': self.username,
            'token': self.site.tokens['csrf'],
            'subject': subject,
            'text': text,
        }
        if ccme:
            params['ccme'] = 1
        mailrequest = self.site.simple_request(**params)
        maildata = mailrequest.submit()

        return ('emailuser' in maildata
                and maildata['emailuser']['result'] == 'Success')

    def block(self, *args, **kwargs) -> None:
        """Block user.

        Refer :py:obj:`APISite.blockuser` method for parameters.
        """
        try:
            self.site.blockuser(self, *args, **kwargs)
        except APIError as err:
            if err.code == 'invalidrange':
                raise ValueError(f'{self.username} is not a valid IP range.')

            raise

    def unblock(self, reason: str | None = None) -> None:
        """Remove the block for the user.

        :param reason: Reason for the unblock.
        """
        self.site.unblockuser(self, reason)

    def logevents(self, **kwargs):
        """Yield user activities.

        :keyword logtype: only iterate entries of this type (see
            mediawiki api documentation for available types)
        :type logtype: str
        :keyword page: only iterate entries affecting this page
        :type page: Page or str
        :keyword namespace: namespace to retrieve logevents from
        :type namespace: int or Namespace
        :keyword start: only iterate entries from and after this
            Timestamp
        :type start: Timestamp or ISO date string
        :keyword end: only iterate entries up to and through this
            Timestamp
        :type end: Timestamp or ISO date string
        :keyword reverse: if True, iterate oldest entries first
            (default: newest)
        :type reverse: bool
        :keyword tag: only iterate entries tagged with this tag
        :type tag: str
        :keyword total: maximum number of events to iterate
        :type total: int
        :rtype: iterable
        """
        return self.site.logevents(user=self.username, **kwargs)

    @property
    def last_event(self):
        """Return last user activity.

        :return: last user log entry
        :rtype: LogEntry or None
        """
        return next(self.logevents(total=1), None)

    def contributions(
        self,
        total: int | None = 500,
        **kwargs
    ) -> Generator[
        tuple[Page, int, pywikibot.Timestamp, str | None],
        None,
        None
    ]:
        """Yield tuples describing this user edits.

        Each tuple is composed of a pywikibot.Page object, the revision
        id, the edit timestamp and the comment. Pages returned are not
        guaranteed to be unique.

        Example:

        >>> site = pywikibot.Site('wikipedia:test')
        >>> user = pywikibot.User(site, 'pywikibot-test')
        >>> contrib = next(user.contributions(reverse=True))
        >>> len(contrib)
        4
        >>> contrib[0].title()
        'User:John Vandenberg/appendtext test'
        >>> contrib[1]
        504588
        >>> str(contrib[2])
        '2022-03-04T17:36:02Z'
        >>> contrib[3]
        ''

        .. seealso:: :meth:`Site.usercontribs()
           <pywikibot.site._generators.GeneratorsMixin.usercontribs>`

        :param total: limit result to this number of pages
        :keyword start: Iterate contributions starting at this Timestamp
        :keyword end: Iterate contributions ending at this Timestamp
        :keyword reverse: Iterate oldest contributions first (default: newest)
        :keyword namespaces: only iterate pages in these namespaces
        :type namespaces: iterable of str or Namespace key,
            or a single instance of those types. May be a '|' separated
            list of namespace identifiers.
        :keyword showMinor: if True, iterate only minor edits; if False and
            not None, iterate only non-minor edits (default: iterate both)
        :keyword top_only: if True, iterate only edits which are the latest
            revision (default: False)
        :return: tuple of pywikibot.Page, revid, pywikibot.Timestamp, comment
        """
        for contrib in self.site.usercontribs(
                user=self.username, total=total, **kwargs):
            ts = pywikibot.Timestamp.fromISOformat(contrib['timestamp'])
            yield (Page(self.site, contrib['title'], contrib['ns']),
                   contrib['revid'],
                   ts,
                   contrib.get('comment'))

    @property
    def first_edit(
        self
    ) -> tuple[Page, int, pywikibot.Timestamp, str | None] | None:
        """Return first user contribution.

        :return: first user contribution entry
        :return: tuple of pywikibot.Page, revid, pywikibot.Timestamp,
            comment
        """
        return next(self.contributions(reverse=True, total=1), None)

    @property
    def last_edit(
        self
    ) -> tuple[Page, int, pywikibot.Timestamp, str | None] | None:
        """Return last user contribution.

        :return: last user contribution entry
        :return: tuple of pywikibot.Page, revid, pywikibot.Timestamp,
            comment
        """
        return next(self.contributions(total=1), None)

    def deleted_contributions(
        self,
        *,
        total: int | None = 500,
        **kwargs,
    ) -> Generator[tuple[Page, Revision], None, None]:
        """Yield tuples describing this user's deleted edits.

        .. versionadded:: 5.5

        :param total: Limit results to this number of pages
        :keyword start: Iterate contributions starting at this Timestamp
        :keyword end: Iterate contributions ending at this Timestamp
        :keyword reverse: Iterate oldest contributions first (default: newest)
        :keyword namespaces: Only iterate pages in these namespaces
        """
        for data in self.site.alldeletedrevisions(user=self.username,
                                                  total=total, **kwargs):
            page = Page(self.site, data['title'], data['ns'])
            for contrib in data['revisions']:
                yield page, Revision(**contrib)

    def uploadedImages(self, total: int = 10):  # noqa: N802
        """Yield tuples describing files uploaded by this user.

        Each tuple is composed of a pywikibot.Page, the timestamp (str
        in ISO8601 format), comment (str) and a bool for pageid > 0.
        Pages returned are not guaranteed to be unique.

        :param total: limit result to this number of pages
        """
        if not self.isRegistered():
            return
        for item in self.logevents(logtype='upload', total=total):
            yield (item.page(),
                   str(item.timestamp()),
                   item.comment(),
                   item.pageid() > 0)

    @property
    def is_thankable(self) -> bool:
        """Determine if the user has thanks notifications enabled.

        .. note::
           This doesn't accurately determine if thanks is enabled for user.
           Privacy of thanks preferences is under discussion, please see
           :phab:`T57401#2216861` and :phab:`T120753#1863894`.
        """
        return self.isRegistered() and 'bot' not in self.groups()

    def renamed_target(self) -> User:
        """Return a User object for the target this user was renamed to.

        If this user was not renamed, it will raise a
        :exc:`NoRenameTargetError`.

        **Usage:**

        >>> site = pywikibot.Site('wikipedia:de')
        >>> user = pywikibot.User(site, 'Foo')
        >>> user.isRegistered()
        False
        >>> target = user.renamed_target()
        >>> target.isRegistered()
        True
        >>> target.title(with_ns=False)
        'Foo~dewiki'
        >>> target.renamed_target()
        Traceback (most recent call last):
        ...
        pywikibot.exceptions.NoRenameTargetError: Rename target user ...

        .. seealso::
           * :meth:`BasePage.moved_target`
           * :meth:`BasePage.getRedirectTarget`

        .. versionadded:: 9.4

        :raises NoRenameTargetError: user was not renamed
        """
        gen = iter(self.site.logevents(logtype='renameuser',
                                       page=self, total=1))
        try:
            renamed = next(gen)
        except StopIteration:
            raise NoRenameTargetError(self)

        return User(self.site, renamed.params['newuser'])



"""Page filter generators provided by the pagegenerators module."""
#
# (C) Pywikibot team, 2008-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import calendar
import io
import re
import typing
from collections import abc
from functools import partial
from http import HTTPStatus
from typing import TYPE_CHECKING, Any
from urllib.parse import urlparse

from requests.exceptions import ReadTimeout

import pywikibot
from pywikibot import config, date, xmlreader
from pywikibot.backports import (
    Callable,
    Generator,
    Iterable,
    Iterator,
    Sequence,
    batched,
)
from pywikibot.comms import http
from pywikibot.exceptions import APIError, ServerError
from pywikibot.site import Namespace
from pywikibot.tools import issue_deprecation_warning
from pywikibot.tools.collections import GeneratorWrapper
from pywikibot.tools.itertools import filter_unique


if TYPE_CHECKING:
    from pywikibot.site import BaseSite, NamespaceArgType
    from pywikibot.site._namespace import SingleNamespaceType
    from pywikibot.time import Timestamp

# This is the function that will be used to de-duplicate page iterators.
_filter_unique_pages = partial(
    filter_unique, key=lambda page: '{}:{}:{}'.format(*page._cmpkey()))


def AllpagesPageGenerator(
    start: str = '!',
    namespace: SingleNamespaceType = 0,
    includeredirects: typing.Literal['only'] | bool = True,
    site: BaseSite | None = None,
    total: int | None = None,
    content: bool = False,
    *,
    filterredir: bool | None = None,
) -> Iterable[pywikibot.page.Page]:
    """Iterate Page objects for all titles in a single namespace.

    .. deprecated:: 10.0
       The *includeredirects* parameter; use *filterredir* instead.
    .. seealso:: :meth:`APISite.allpages()
       <pywikibot.site._generators.GeneratorsMixin.allpages>`

    :param start: if provided, only generate pages >= this title
        lexically
    :param namespace: Namespace to retrieve pages from
    :param includeredirects: If False, redirects are not included. If
        equals the string 'only', only redirects are added. Otherwise
        redirects will be included. This parameter is deprecated; use
        *filterredir* instead.
    :param site: Site for generator results.
    :param total: Maximum number of pages to retrieve in total
    :param content: If True, load current version of each page (default
        False)
    :param filterredir: if True, only yield redirects; if False (and
        not None), only yield non-redirects (default: yield both).
    :return: a generator that yields Page objects
    :raises ValueError: *filterredir* as well as *includeredirects*
        parameters were given. Use *filterredir* only.
    """
    if site is None:
        site = pywikibot.Site()

    if filterredir is not None and includeredirects is not True:
        raise ValueError(
            f'filterredir parameter ({filterredir}) is used together with '
            f'outdated includeredirects parameter ({includeredirects}).'
        )

    # backward compatibility
    if includeredirects is not True:
        if not includeredirects:
            filterredir = False
        elif includeredirects == 'only':
            filterredir = True

        issue_deprecation_warning(
            'includeredirects parameter ({includeredirects})',
            f'filterredir={filterredir}',
            since='10.0.0'
        )

    return site.allpages(start=start, namespace=namespace,
                         filterredir=filterredir, total=total, content=content)


def PrefixingPageGenerator(
    prefix: str,
    namespace: SingleNamespaceType | None = None,
    includeredirects: typing.Literal['only'] | bool = True,
    site: BaseSite | None = None,
    total: int | None = None,
    content: bool = False,
    *,
    filterredir: bool | None = None,
) -> Iterable[pywikibot.page.Page]:
    """Prefixed Page generator.

    .. deprecated:: 10.0
       The *includeredirects* parameter; use *filterredir* instead.

    :param prefix: The prefix of the pages.
    :param namespace: Namespace to retrieve pages from
    :param includeredirects: If False, redirects are not included. If
        equals the string 'only', only redirects are added. Otherwise
        redirects will be included. This parameter is deprecated; use
        *filterredir* instead.
    :param site: Site for generator results.
    :param total: Maximum number of pages to retrieve in total
    :param content: If True, load current version of each page (default
        False)
    :param filterredir: if True, only yield redirects; if False (and
        not None), only yield non-redirects (default: yield both).
    :return: a generator that yields Page objects
    :raises ValueError: *filterredir* as well as *includeredirects*
        parameters were given. Use *filterredir* only.
    """
    if site is None:
        site = pywikibot.Site()

    prefixlink = pywikibot.Link(prefix, site)
    if namespace is None:
        namespace = prefixlink.namespace
    title = prefixlink.title

    if filterredir is not None and includeredirects is not True:
        raise ValueError(
            f'filterredir parameter ({filterredir}) is used together with '
            f'outdated includeredirects parameter ({includeredirects}).'
        )

    # backward compatibility
    if includeredirects is not True:
        if not includeredirects:
            filterredir = False
        elif includeredirects == 'only':
            filterredir = True

        issue_deprecation_warning(
            'includeredirects parameter ({includeredirects})',
            f'filterredir={filterredir}',
            since='10.0.0'
        )

    return site.allpages(prefix=title, namespace=namespace,
                         filterredir=filterredir, total=total, content=content)


def LogeventsPageGenerator(logtype: str | None = None,
                           user: str | None = None,
                           site: BaseSite | None = None,
                           namespace: SingleNamespaceType | None = None,
                           total: int | None = None,
                           start: Timestamp | None = None,
                           end: Timestamp | None = None,
                           reverse: bool = False
                           ) -> Generator[pywikibot.page.Page, None, None]:
    """Generate Pages for specified modes of logevents.

    :param logtype: Mode of logs to retrieve
    :param user: User of logs retrieved
    :param site: Site for generator results
    :param namespace: Namespace to retrieve logs from
    :param total: Maximum number of pages to retrieve in total
    :param start: Timestamp to start listing from
    :param end: Timestamp to end listing at
    :param reverse: if True, start with oldest changes (default: newest)
    """
    if site is None:
        site = pywikibot.Site()
    for entry in site.logevents(total=total, logtype=logtype, user=user,
                                namespace=namespace, start=start, end=end,
                                reverse=reverse):
        try:
            yield entry.page()
        except KeyError as e:
            pywikibot.warning('LogeventsPageGenerator: failed to load page '
                              f'for {entry.data!r}; skipping')
            pywikibot.error(e)


def NewpagesPageGenerator(site: BaseSite | None = None,
                          namespaces: NamespaceArgType = (0, ),
                          total: int | None = None
                          ) -> Generator[pywikibot.page.Page, None, None]:
    """Iterate Page objects for all new titles in a single namespace.

    :param site: Site for generator results.
    :param namespaces: namespace to retrieve pages from
    :param total: Maximum number of pages to retrieve in total
    """
    # API does not (yet) have a newpages function, so this tries to duplicate
    # it by filtering the recentchanges output
    # defaults to namespace 0 because that's how Special:Newpages defaults
    if site is None:
        site = pywikibot.Site()
    return (page for page, _ in site.newpages(namespaces=namespaces,
                                              total=total, returndict=True))


def RecentChangesPageGenerator(
    site: BaseSite | None = None,
    _filter_unique: None | (Callable[[Iterable[pywikibot.Page]],
                            Iterable[pywikibot.Page]]) = None,
    **kwargs: Any
) -> Generator[pywikibot.Page, None, None]:
    """Generate recent changes pages, including duplicates.

    For keyword parameters refer :meth:`APISite.recentchanges()
    <pywikibot.site._generators.GeneratorsMixin.recentchanges>`.

    .. versionchanged:: 8.2
       The YieldType depends on namespace. It can be
       :class:`pywikibot.Page<pywikibot.page.Page>`,
       :class:`pywikibot.User<pywikibot.page.User>`,
       :class:`pywikibot.FilePage<pywikibot.page.FilePage>` or
       :class:`pywikibot.Category<pywikibot.page.Category>`.
    .. versionchanged:: 9.4
       Ignore :class:`pywikibot.FilePage<pywikibot.page.FilePage>` if it
       raises a :exc:`ValueError` during upcast e.g. due to an invalid
       file extension.

    :param site: Site for generator results.
    """
    def upcast(gen):
        """Upcast pywikibot.Page type."""
        for rc in gen:
            # The title in a log entry may have been suppressed
            if rc['type'] == 'log' and 'title' not in rc:
                continue

            ns = rc['ns']
            if ns == Namespace.USER:
                pageclass: type[pywikibot.Page] = pywikibot.User
            elif ns == Namespace.FILE:
                pageclass = pywikibot.FilePage
            elif ns == Namespace.CATEGORY:
                pageclass = pywikibot.Category
            else:
                pageclass = pywikibot.Page
            try:
                yield pageclass(site, rc['title'])
            except ValueError:
                if pageclass != pywikibot.FilePage:
                    raise

                pywikibot.exception()

    if site is None:
        site = pywikibot.Site()

    gen = site.recentchanges(**kwargs)
    gen.request['rcprop'] = 'title'
    gen = upcast(gen)

    if _filter_unique:
        gen = _filter_unique(gen)
    return gen


def UnconnectedPageGenerator(
    site: BaseSite | None = None,
    total: int | None = None,
    *,
    strict: bool = False
) -> Iterable[pywikibot.page.Page]:
    """Iterate Page objects for all unconnected pages to a Wikibase repository.

    .. versionchanged::
       The *strict* parameter was added.

    :param site: Site for generator results.
    :param total: Maximum number of pages to retrieve in total
    :param strict: If ``True``, verify that each page still has no data
        item before yielding it.
    :raises ValueError: The given site does not have Wikibase repository
    """
    if site is None:
        site = pywikibot.Site()
    if not site.data_repository():
        raise ValueError('The given site does not have Wikibase repository.')
    return site.unconnected_pages(total=total, strict=strict)


def FileLinksGenerator(
    referredFilePage: pywikibot.page.FilePage,  # noqa: N803
    total: int | None = None,
    content: bool = False
) -> Iterable[pywikibot.page.Page]:
    """Yield Pages on which referredFilePage file is displayed."""
    return referredFilePage.using_pages(total=total, content=content)


def ImagesPageGenerator(
    pageWithImages: pywikibot.page.Page,  # noqa: N803
    total: int | None = None,
    content: bool = False
) -> Iterable[pywikibot.page.Page]:
    """Yield FilePages displayed on pageWithImages."""
    return pageWithImages.imagelinks(total=total, content=content)


def InterwikiPageGenerator(page: pywikibot.page.Page
                           ) -> Generator[pywikibot.page.Page, None, None]:
    """Iterate over all interwiki (non-language) links on a page."""
    return (pywikibot.Page(link) for link in page.interwiki())


def LanguageLinksPageGenerator(page: pywikibot.page.Page,
                               total: int | None = None
                               ) -> Generator[pywikibot.page.Page, None, None]:
    """Iterate over all interwiki language links on a page."""
    return (pywikibot.Page(link) for link in page.iterlanglinks(total=total))


def CategorizedPageGenerator(category: pywikibot.page.Category,
                             recurse: int | bool = False,
                             start: str | None = None,
                             total: int | None = None,
                             content: bool = False,
                             namespaces: NamespaceArgType = None,
                             ) -> Generator[pywikibot.page.Page, None, None]:
    """Yield all pages in a specific category.

    :param recurse: if not False or 0, also iterate articles in
        subcategories. If an int, limit recursion to this number of
        levels. (Example: recurse=1 will iterate articles in first-level
        subcats, but no deeper.)
    :param start: if provided, only generate pages >= this title
        lexically
    :param total: iterate no more than this number of pages in total (at
        all levels)
    :param content: if True, retrieve the content of the current version
        of each page (default False)
    """
    yield from category.articles(
        content=content,
        namespaces=namespaces,
        recurse=recurse,
        startprefix=start,
        total=total,
    )


def SubCategoriesPageGenerator(category: pywikibot.page.Category,
                               recurse: int | bool = False,
                               start: str | None = None,
                               total: int | None = None,
                               content: bool = False,
                               ) -> Generator[pywikibot.page.Page, None, None]:
    """Yield all subcategories in a specific category.

    :param recurse: if not False or 0, also iterate articles in
        subcategories. If an int, limit recursion to this number of
        levels. (Example: recurse=1 will iterate articles in first-level
        subcats, but no deeper.)
    :param start: if provided, only generate pages >= this title
        lexically
    :param total: iterate no more than this number of pages in total (at
        all levels)
    :param content: if True, retrieve the content of the current version
        of each page (default False)
    """
    # TODO: page generator could be modified to use cmstartsortkey ...
    for s in category.subcategories(recurse=recurse,
                                    total=total, content=content):
        if start is None or s.title(with_ns=False) >= start:
            yield s


def LinkedPageGenerator(
    linkingPage: pywikibot.page.Page,  # noqa: N803
    total: int | None = None,
    content: bool = False
) -> Iterable[pywikibot.page.BasePage]:
    """Yield all pages linked from a specific page.

    See :py:obj:`page.BasePage.linkedPages` for details.

    :param linkingPage: the page that links to the pages we want
    :param total: the total number of pages to iterate
    :param content: if True, retrieve the current content of each linked page
    :return: a generator that yields Page objects of pages linked to
        linkingPage
    """
    return linkingPage.linkedPages(total=total, content=content)


def _yield_titles(f: io.TextIOBase,
                  site: pywikibot.site.BaseSite
                  ) -> Generator[pywikibot.page.Page, None, None]:
    """Yield page titles from a text stream.

    :param f: text stream object
    :param site: Site for generator results.
    :return: a generator that yields Page objects of pages with titles
        in text stream
    """
    linkmatch = None
    for linkmatch in pywikibot.link_regex.finditer(f.read()):
        # If the link is in interwiki format, the Page object may reside
        # on a different Site than the default.
        # This makes it possible to work on different wikis using a single
        # text file, but also could be dangerous because you might
        # inadvertently change pages on another wiki!
        yield pywikibot.Page(pywikibot.Link(linkmatch['title'], site))

    if linkmatch is not None:
        return

    f.seek(0)
    for title in f:
        title = title.strip()
        if '|' in title:
            title = title[:title.index('|')]
        if title:
            yield pywikibot.Page(site, title)


def TextIOPageGenerator(source: str | None = None,
                        site: BaseSite | None = None,
                        ) -> Generator[pywikibot.page.Page, None, None]:
    """Iterate pages from a list in a text file or on a webpage.

    The text source must contain page links between double-square-
    brackets or, alternatively, separated by newlines. The generator
    will yield each corresponding Page object.

    :param source: the file path or URL that should be read. If no name
        is given, the generator prompts the user.
    :param site: Site for generator results.
    """
    if source is None:
        source = pywikibot.input('Please enter the filename / URL:')
    if site is None:
        site = pywikibot.Site()
    # If source cannot be parsed as an HTTP URL, treat as local file
    if not urlparse(source).netloc:
        with open(source, encoding=config.textfile_encoding) as local_file:
            yield from _yield_titles(local_file, site)
    # Else, fetch page (page should return text in same format as that expected
    # in filename, i.e. pages separated by newlines or pages enclosed in double
    # brackets
    else:
        with io.StringIO(http.fetch(source).text) as f:
            yield from _yield_titles(f, site)


def PagesFromTitlesGenerator(iterable: Iterable[str],
                             site: BaseSite | None = None
                             ) -> Generator[pywikibot.page.Page, None, None]:
    """Generate pages from the titles (strings) yielded by iterable.

    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    for title in iterable:
        if not isinstance(title, str):
            break
        yield pywikibot.Page(pywikibot.Link(title, site))


def PagesFromPageidGenerator(
    pageids: Iterable[str],
    site: BaseSite | None = None
) -> Iterable[pywikibot.page.Page]:
    """Return a page generator from pageids.

    Pages are iterated in the same order than in the underlying pageids.
    Pageids are filtered and only one page is returned in case of
    duplicate pageid.

    :param pageids: an iterable that returns pageids, or a comma-
        separated string of pageids (e.g. '945097,1483753,956608')
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()

    return site.load_pages_from_pageids(pageids)


def UserContributionsGenerator(username: str,
                               namespaces: NamespaceArgType = None,
                               site: BaseSite | None = None,
                               total: int | None = None,
                               _filter_unique: None | (Callable[
                                   [Iterable[pywikibot.page.Page]],
                                   Iterable[pywikibot.page.Page]]) =
                               _filter_unique_pages
                               ) -> Iterable[pywikibot.page.Page]:
    """Yield unique pages edited by user:username.

    :param total: Maximum number of pages to retrieve in total
    :param namespaces: list of namespace numbers to fetch contribs from
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()

    user = pywikibot.User(site, username)
    if not (user.isAnonymous() or user.isRegistered()):
        pywikibot.warning(
            f'User "{user.username}" does not exist on site "{site}".')

    gen = (contrib[0] for contrib in user.contributions(
        namespaces=namespaces, total=total))
    if _filter_unique:
        return _filter_unique(gen)
    return gen


def NewimagesPageGenerator(total: int | None = None,
                           site: BaseSite | None = None
                           ) -> Generator[pywikibot.page.Page, None, None]:
    """New file generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return (entry.page()
            for entry in site.logevents(logtype='upload', total=total))


def WikibaseItemGenerator(gen: Iterable[pywikibot.page.Page]
                          ) -> Generator[pywikibot.page.ItemPage, None, None]:
    """A wrapper generator used to yield Wikibase items of another generator.

    :param gen: Generator to wrap.
    :return: Wrapped generator
    """
    for page in gen:
        if isinstance(page, pywikibot.ItemPage):
            yield page
        elif page.site.data_repository() == page.site:
            # These are already items, as they have a DataSite in page.site.
            # However generator is yielding Page, so convert to ItemPage.
            # FIXME: If we've already fetched content, we should retain it
            yield pywikibot.ItemPage(page.site, page.title())
        else:
            yield pywikibot.ItemPage.fromPage(page)


def AncientPagesPageGenerator(
    total: int = 100,
    site: BaseSite | None = None
) -> Generator[pywikibot.page.Page, None, None]:
    """Ancient page generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return (page for page, _ in site.ancientpages(total=total))


def UnusedFilesGenerator(
    total: int | None = None,
    site: BaseSite | None = None
) -> Iterable[pywikibot.page.FilePage]:
    """Unused files generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return site.unusedfiles(total=total)


def WithoutInterwikiPageGenerator(
    total: int | None = None,
    site: BaseSite | None = None
) -> Iterable[pywikibot.page.Page]:
    """Page lacking interwikis generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return site.withoutinterwiki(total=total)


def UnCategorizedCategoryGenerator(
    total: int | None = 100,
    site: BaseSite | None = None
) -> Iterable[pywikibot.Category]:
    """Uncategorized category generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return site.uncategorizedcategories(total=total)


def UnCategorizedImageGenerator(
    total: int = 100,
    site: BaseSite | None = None
) -> Iterable[pywikibot.page.FilePage]:
    """Uncategorized file generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return site.uncategorizedimages(total=total)


def UnCategorizedPageGenerator(
    total: int = 100,
    site: BaseSite | None = None
) -> Iterable[pywikibot.page.Page]:
    """Uncategorized page generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return site.uncategorizedpages(total=total)


def UnCategorizedTemplateGenerator(
    total: int = 100,
    site: BaseSite | None = None
) -> Iterable[pywikibot.page.Page]:
    """Uncategorized template generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return site.uncategorizedtemplates(total=total)


def LonelyPagesPageGenerator(
    total: int | None = None,
    site: BaseSite | None = None
) -> Iterable[pywikibot.page.Page]:
    """Lonely page generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return site.lonelypages(total=total)


def UnwatchedPagesPageGenerator(
    total: int | None = None,
    site: BaseSite | None = None
) -> Iterable[pywikibot.page.Page]:
    """Unwatched page generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return site.unwatchedpages(total=total)


def page_with_property_generator(
    name: str,
    total: int | None = None,
    site: BaseSite | None = None
) -> Iterable[pywikibot.page.Page]:
    """Special:PagesWithProperty page generator.

    :param name: Property name of pages to be retrieved
    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return site.pages_with_property(name, total=total)


def WantedPagesPageGenerator(
    total: int = 100,
    site: BaseSite | None = None
) -> Iterable[pywikibot.page.Page]:
    """Wanted page generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return site.wantedpages(total=total)


def DeadendPagesPageGenerator(
    total: int = 100,
    site: BaseSite | None = None
) -> Iterable[pywikibot.page.Page]:
    """Dead-end page generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return site.deadendpages(total=total)


def LongPagesPageGenerator(total: int = 100,
                           site: BaseSite | None = None
                           ) -> Generator[pywikibot.page.Page, None, None]:
    """Long page generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return (page for page, _ in site.longpages(total=total))


def ShortPagesPageGenerator(total: int = 100,
                            site: BaseSite | None = None
                            ) -> Generator[pywikibot.page.Page, None, None]:
    """Short page generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return (page for page, _ in site.shortpages(total=total))


def RandomPageGenerator(
    total: int | None = None,
    site: BaseSite | None = None,
    namespaces: NamespaceArgType = None
) -> Iterable[pywikibot.page.Page]:
    """Random page generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return site.randompages(total=total, namespaces=namespaces)


def RandomRedirectPageGenerator(
    total: int | None = None,
    site: BaseSite | None = None,
    namespaces: NamespaceArgType = None,
) -> Iterable[pywikibot.page.Page]:
    """Random redirect generator.

    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    return site.randompages(total=total, namespaces=namespaces,
                            redirects=True)


def LinksearchPageGenerator(
    url: str,
    namespaces: NamespaceArgType = None,
    total: int | None = None,
    site: BaseSite | None = None,
    protocol: str | None = None
) -> Iterable[pywikibot.page.Page]:
    """Yield all pages that link to a certain URL.

    :param url: The URL to search for (with or without the protocol
        prefix); this may include a '*' as a wildcard, only at the start
        of the hostname
    :param namespaces: list of namespace numbers to fetch contribs from
    :param total: Maximum number of pages to retrieve in total
    :param site: Site for generator results
    :param protocol: Protocol to search for, likely http or https, http
        by default. Full list shown on Special:LinkSearch wikipage.
    """
    if site is None:
        site = pywikibot.Site()
    return site.exturlusage(url, namespaces=namespaces, protocol=protocol,
                            total=total, content=False)


def SearchPageGenerator(
    query: str,
    total: int | None = None,
    namespaces: NamespaceArgType = None,
    site: BaseSite | None = None,
    **kwargs
) -> Iterable[pywikibot.page.Page]:
    r"""Yield pages from the MediaWiki internal search engine.

    .. versionchanged:: 10.0
       Keyword arguments *content*, *sort* and *where* was added.

    .. seealso:: :meth:`site.search()
       <pywikibot.site._generators.GeneratorsMixin.search>`

    :param query: the text to search for
    :param total: Maximum number of pages to retrieve in total
    :param namespaces: search only in these namespaces (defaults to all)
    :param site: Site for generator results.
    :keyword str | None where: Where to search; value must be one of the
        given literals or None (many wikis do not support all search
        types)
    :keyword bool content: if True, load the current content of each
        iterated page (default False)
    :keyword sort: Set the sort order of returned results. If None is
        given, 'none' is used. Default is sort by relevance.
    """
    if site is None:
        site = pywikibot.Site()
    return site.search(query, total=total, namespaces=namespaces, **kwargs)


def LiveRCPageGenerator(site: BaseSite | None = None,
                        total: int | None = None
                        ) -> Generator[pywikibot.page.Page, None, None]:
    """Yield pages from a socket.io RC stream.

    Generates pages based on the EventStreams Server-Sent-Event (SSE) recent
    changes stream.
    The Page objects will have an extra property ._rcinfo containing the
    literal rc data. This can be used to e.g. filter only new pages. See
    `pywikibot.comms.eventstreams.rc_listener` for details on the .rcinfo
    format.

    :param site: site to return recent changes for
    :param total: the maximum number of changes to return
    """
    if site is None:
        site = pywikibot.Site()

    from pywikibot.comms.eventstreams import site_rc_listener

    for entry in site_rc_listener(site, total=total):
        # The title in a log entry may have been suppressed
        if 'title' not in entry and entry['type'] == 'log':
            continue
        page = pywikibot.Page(site, entry['title'], entry['namespace'])
        page._rcinfo = entry  # type: ignore[attr-defined]
        yield page


class GoogleSearchPageGenerator(GeneratorWrapper):

    """Page generator using Google search results.

    To use this generator, you need to install the googlesearch package::

        pip install googlesearch-python

    As there are concerns about Google's Terms of Service, this
    generator prints a warning for each query.

    .. seealso:: https://policies.google.com/terms
    .. versionchanged:: 7.6
       subclassed from :class:`tools.collections.GeneratorWrapper`
    .. versionchanged:: 10.1
       ``googlesearch-python`` package is needed instead of ``google``,
       see :phab:`T387618` for further informations. The *total*
       parameter was added. The *query* parameter is positional only.
       All other parameters are keyword only.
    """

    def __init__(self, query: str = '', /, *,
                 site: BaseSite | None = None,
                 total: int = 10) -> None:
        """Initializer.

        :param query: the text to search for.
        :param site: Site for generator results.
        :param total: the maximum number of changes to return, default
            is 10 which is also set by googlesearch package.
        """
        self.query = query or pywikibot.input(
            'Please enter the search query:')
        self.site = site or pywikibot.Site()
        self.limit = total

    @staticmethod
    def queryGoogle(query: str, /, **kwargs) -> Generator[str, None, None]:
        """Perform a query using ``googlesearch-python`` package.

        .. admonition:: Terms of Service

           The terms of service as at June 2014 give two conditions that
           may apply to use of search:

           1. Don't access [Google Services] using a method other than
              the interface and the instructions that [they] provide.
           2. Don't remove, obscure, or alter any legal notices
              displayed in or along with [Google] Services.

           Both of those issues should be managed by the
           ``googlesearch-python`` package, however Pywikibot will at
           least ensure the user sees the TOS in order to comply with
           the second.

           .. seealso:: https://policies.google.com/terms
              condition.

        .. important:: These note are from 2014 and have not been
           reviewed or updated since then.

        .. versionchanged:: 10.1
           *query* is positional only; *kwargs* parameter was added.

        :param query: the text to search for.
        :param kwargs: other keyword arguments passed to ``googlesearch``
            module.
        """
        try:
            import googlesearch
        except ModuleNotFoundError:
            pywikibot.error("""\
generator GoogleSearchPageGenerator depends on package
'googlesearch-python'. To install, please run:

    pip install googlesearch-python""")
            return

        pywikibot.warning('Please read http://www.google.com/accounts/TOS')
        yield from googlesearch.search(query, **kwargs)

    @property
    def generator(self) -> Generator[pywikibot.page.Page, None, None]:
        """Yield results from :meth:`queryGoogle` query.

        .. versionchanged:: 7.6
           changed from iterator method to generator property
        .. versionchanged:: 10.1
           use :meth:`site.protocol
           <pywikibot.site._basesite.BaseSite.protocol>` to get the base
           URL. Also filter duplicates.
        """
        if not self.query:
            pywikibot.warning('No query string was specified')
            return

        # restrict query to local site
        site = self.site
        local_query = f'{self.query} site:{site.hostname()}'
        base = f'{site.protocol()}://{site.hostname()}{site.articlepath}'
        pattern = re.compile(base.replace('{}', '(?P<title>.+)'))

        for url in self.queryGoogle(local_query, num_results=self.limit,
                                    unique=True):
            m = pattern.fullmatch(url)
            if not m:
                continue

            page = pywikibot.Page(pywikibot.Link(m['title'], site))

            # Google may contain links in the format:
            # https://de.wikipedia.org/wiki/en:Foobar
            if page.site == site:
                yield page

    def set_maximum_items(self, value: int, /) -> None:
        """Set the maximum number of items to be retrieved from google.

        This method is added to be used by the
        :class:`pagegenerators.GeneratorFactory` to circumvent call of
        :func:`itertools.islice` filter for this generator.

        .. versionadded:: 10.1
        """
        self.limit = value


def MySQLPageGenerator(query: str, site: BaseSite | None = None,
                       verbose: bool | None = None
                       ) -> Generator[pywikibot.page.Page, None, None]:
    """Yield a list of pages based on a MySQL query.

    The query should return two columns, page namespace and page title pairs
    from some table. An example query that yields all ns0 pages might look
    like::

        SELECT
         page_namespace,
         page_title
        FROM page
        WHERE page_namespace = 0;

    .. seealso:: :manpage:`MySQL`

    :param query: MySQL query to execute
    :param site: Site object
    :param verbose: if True, print query to be executed;
        if None, config.verbose_output will be used.
    :return: generator which yields pywikibot.Page
    """
    from pywikibot.data import mysql

    if site is None:
        site = pywikibot.Site()

    row_gen = mysql.mysql_query(query,
                                dbname=site.dbName(),
                                verbose=verbose)

    for row in row_gen:
        namespace_number, page_name = row
        page_name = page_name.decode(site.encoding())
        page = pywikibot.Page(site, page_name, ns=int(namespace_number))
        yield page


def SupersetPageGenerator(query: str,
                          site: BaseSite | None = None,
                          schema_name: str | None = None,
                          database_id: int | None = None
                          ) -> Iterator[pywikibot.page.Page]:
    """Generate pages that result from the given SPARQL query.

    Pages are generated using site in following order:

    1. site retrieved using page_wikidb column in SQL result
    2. site as parameter
    3. site retrieved using schema_name

    SQL columns used are

    - page_id
    - page_namespace + page_title
    - page_wikidb

    Example SQL queries

    .. code-block:: sql

        SELECT
            gil_wiki AS page_wikidb,
            gil_page AS page_id
        FROM globalimagelinks
        GROUP BY gil_wiki
        LIMIT 10

    OR

    .. code-block:: sql

        SELECT
            page_id
        FROM page
        LIMIT 10

    OR

    .. code-block:: sql

        SELECT
            page_namespace,
            page_title
        FROM page
        LIMIT 10

    .. versionadded:: 9.2

    :param query: the SQL query string.
    :param site: Site for generator results.
    :param schema_name: target superset schema name
    :param database_id: target superset database id
    """
    from pywikibot.data.superset import SupersetQuery

    # Do not pass site to superset if schema_name is defined.
    # The user may use schema_name to point to different
    # wikimedia db on purpose and use site for
    # generating result pages.

    superset_site = None if schema_name else site

    superset = SupersetQuery(site=superset_site,
                             schema_name=schema_name,
                             database_id=database_id)

    try:
        rows = superset.query(query)
    except Exception as e:
        pywikibot.error(f'Error executing query: {query}\n{e}')
        return

    sites = {}

    # If there is no site then retrieve it using schema_name
    if not site:
        if not schema_name:
            raise TypeError('Schema name or site must be provided.')

        wikidb = re.sub('_p$', '', schema_name)
        site = pywikibot.site.APISite.fromDBName(wikidb)

    for row in rows:
        # If page_wikidb column in SQL result then use it to retrieve site
        if 'page_wikidb' in row:
            # remove "_p" suffix
            wikidb = re.sub('_p$', '', row['page_wikidb'])

            # Caching sites
            if wikidb not in sites:
                try:
                    sites[wikidb] = pywikibot.site.APISite.fromDBName(wikidb)
                except ValueError:
                    msg = f'Cannot parse a site from {wikidb} for {row}.'
                    pywikibot.warning(msg)
                    continue
            site = sites[wikidb]

        # Generate page objects

        # Create page object from page_id
        if 'page_id' in row:
            page_ids = [row['page_id']]
            pages = site.load_pages_from_pageids(page_ids)
            for page in pages:
                yield page

        # Create page object from page_namespace + page_title
        elif 'page_title' in rows[0] and 'page_namespace' in rows[0]:
            page_namespace = int(row['page_namespace'])
            page_title = row['page_title']
            page = pywikibot.Page(site, page_title, ns=page_namespace)
            yield page

        else:
            raise ValueError('The SQL result is in wrong format.')


class XMLDumpPageGenerator(abc.Iterator):  # type: ignore[type-arg]

    """Xml iterator that yields Page objects.

    .. versionadded:: 7.2
       the `content` parameter

    :param filename: filename of XML dump
    :param start: skip entries below that value
    :param namespaces: namespace filter
    :param site: current site for the generator
    :param text_predicate: a callable with entry.text as parameter and boolean
        as result to indicate the generator should return the page or not
    :param content: If True, assign old page content to Page.text

    :ivar skipping: True if start parameter is given, else False
    :ivar parser: holds the xmlreader.XmlDump parse method
    """

    def __init__(
        self,
        filename: str,
        start: str | None = None,
        namespaces: NamespaceArgType = None,
        site: BaseSite | None = None,
        text_predicate: Callable[[str], bool] | None = None,
        content=False,
    ) -> None:
        """Initializer."""
        self.text_predicate = text_predicate
        self.content = content
        self.skipping = bool(start)

        self.start: str | None = None
        if start is not None and self.skipping:
            self.start = start.replace('_', ' ')

        self.site = site or pywikibot.Site()
        if not namespaces:
            self.namespaces = self.site.namespaces
        else:
            self.namespaces = self.site.namespaces.resolve(namespaces)
        dump = xmlreader.XmlDump(filename, on_error=pywikibot.error)
        self.parser = dump.parse()

    def __next__(self) -> pywikibot.page.Page:
        """Get next Page."""
        while True:
            entry = next(self.parser)
            if self.skipping:
                if entry.title < self.start:
                    continue
                self.skipping = False
            page = pywikibot.Page(self.site, entry.title)
            if page.namespace() not in self.namespaces:
                continue
            if not self.text_predicate or self.text_predicate(entry.text):
                if self.content:
                    page.text = entry.text
                return page


def YearPageGenerator(start: int = 1, end: int = 2050,
                      site: BaseSite | None = None
                      ) -> Generator[pywikibot.page.Page, None, None]:
    """Year page generator.

    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    pywikibot.info(f'Starting with year {start}')
    for i in range(start, end + 1):
        if i % 100 == 0:
            pywikibot.info(f'Preparing {i}...')
        # There is no year 0
        if i != 0:
            current_year = date.formatYear(site.lang, i)
            yield pywikibot.Page(pywikibot.Link(current_year, site))


def DayPageGenerator(start_month: int = 1, end_month: int = 12,
                     site: BaseSite | None = None, year: int = 2000
                     ) -> Generator[pywikibot.page.Page, None, None]:
    """Day page generator.

    :param site: Site for generator results.
    :param year: considering leap year.
    """
    if site is None:
        site = pywikibot.Site()
    lang = site.lang
    first_page = pywikibot.Page(site, date.format_date(start_month, 1, lang))
    pywikibot.info(f'Starting with {first_page.title(as_link=True)}')
    for month in range(start_month, end_month + 1):
        for day in range(1, calendar.monthrange(year, month)[1] + 1):
            yield pywikibot.Page(
                pywikibot.Link(date.format_date(month, day, lang), site))


def WikidataPageFromItemGenerator(
    gen: Iterable[pywikibot.page.ItemPage],
    site: pywikibot.site.BaseSite,
) -> Generator[pywikibot.page.Page, None, None]:
    """Generate pages from site based on sitelinks of item pages.

    :param gen: generator of :py:obj:`pywikibot.ItemPage`
    :param site: Site for generator results.
    """
    repo = site.data_repository()
    for batch in batched(gen, 50):
        req = {'ids': [item.id for item in batch],
               'sitefilter': site.dbName(),
               'action': 'wbgetentities',
               'props': 'sitelinks'}

        wbrequest = repo.simple_request(**req)
        wbdata = wbrequest.submit()
        entities = (item for item in wbdata['entities'].values() if
                    'sitelinks' in item and site.dbName() in item['sitelinks'])
        sitelinks = (item['sitelinks'][site.dbName()]['title']
                     for item in entities)
        for sitelink in sitelinks:
            yield pywikibot.Page(site, sitelink)


def WikidataSPARQLPageGenerator(query: str,
                                site: BaseSite | None = None,
                                item_name: str = 'item',
                                endpoint: str | None = None,
                                entity_url: str | None = None,
                                result_type: Any = set
                                ) -> Iterator[pywikibot.page.Page]:
    """Generate pages that result from the given SPARQL query.

    :param query: the SPARQL query string.
    :param site: Site for generator results.
    :param item_name: name of the item in the SPARQL query
    :param endpoint: SPARQL endpoint URL
    :param entity_url: URL prefix for any entities returned in a query.
    :param result_type: type of the iterable in which SPARQL results are
        stored (default set)
    """
    from pywikibot.data import sparql

    if site is None:
        site = pywikibot.Site()
    repo = site.data_repository()
    dependencies = {'endpoint': endpoint, 'entity_url': entity_url}
    if not endpoint or not entity_url:
        dependencies['repo'] = repo
    query_object = sparql.SparqlQuery(**dependencies)  # type: ignore[arg-type]
    data = query_object.get_items(query,
                                  item_name=item_name,
                                  result_type=result_type)
    entities = (repo.get_entity_for_entity_id(entity) for entity in data)
    if isinstance(site, pywikibot.site.DataSite):
        return entities

    return WikidataPageFromItemGenerator(entities, site)


def WikibaseSearchItemPageGenerator(
    text: str,
    language: str | None = None,
    total: int | None = None,
    site: BaseSite | None = None,
) -> Generator[pywikibot.page.ItemPage, None, None]:
    """Generate pages that contain the provided text.

    :param text: Text to look for.
    :param language: Code of the language to search in. If not
        specified, value from pywikibot.config.data_lang is used.
    :param total: Maximum number of pages to retrieve in total, or None
        in case of no limit.
    :param site: Site for generator results.
    """
    if site is None:
        site = pywikibot.Site()
    if language is None:
        language = site.lang
    repo = site.data_repository()

    data = repo.search_entities(text, language, total=total)
    return (pywikibot.ItemPage(repo, item['id']) for item in data)


class PetScanPageGenerator(GeneratorWrapper):

    """Queries PetScan to generate pages.

    .. seealso:: https://petscan.wmflabs.org/
    .. versionadded:: 3.0
    .. versionchanged:: 7.6
       subclassed from :class:`tools.collections.GeneratorWrapper`
    """

    def __init__(
        self,
        categories: Sequence[str],
        subset_combination: bool = True,
        namespaces: Iterable[int | pywikibot.site.Namespace] | None = None,
        site: BaseSite | None = None,
        extra_options: dict[Any, Any] | None = None
    ) -> None:
        """Initializer.

        :param categories: List of category names to retrieve pages from
        :param subset_combination: Combination mode. If True, returns
            the intersection of the results of the categories, else
            returns the union of the results of the categories
        :param namespaces: List of namespaces to search in (default is
            None, meaning all namespaces)
        :param site: Site to operate on (default is the default site
            from the user config)
        :param extra_options: Dictionary of extra options to use
            (optional)
        """
        if site is None:
            site = pywikibot.Site()

        self.site = site
        self.opts = self.buildQuery(categories, subset_combination,
                                    namespaces, extra_options)

    def buildQuery(self, categories: Sequence[str], subset_combination: bool,
                   namespaces: Iterable[int | pywikibot.site.Namespace] | None,
                   extra_options: dict[Any, Any] | None) -> dict[str, Any]:
        """Get the querystring options to query PetScan.

        :param categories: List of categories (as strings)
        :param subset_combination: Combination mode. If True, returns
            the intersection of the results of the categories, else
            returns the union of the results of the categories
        :param namespaces: List of namespaces to search in
        :param extra_options: Dictionary of extra options to use
        :return: Dictionary of querystring parameters to use in the
            query
        """
        extra_options = extra_options or {}

        query = {
            'language': self.site.code,
            'project': self.site.hostname().split('.')[-2],
            'combination': 'subset' if subset_combination else 'union',
            'categories': '\r\n'.join(categories),
            'format': 'json',
            'doit': ''
        }

        if namespaces:
            for namespace in namespaces:
                query[f'ns[{int(namespace)}]'] = 1

        query_final = query.copy()
        query_final.update(extra_options)

        return query_final

    def query(self) -> Generator[dict[str, Any], None, None]:
        """Query PetScan.

        .. versionchanged:: 7.4
           raises :class:`APIError` if query returns an error message.

        :raises ServerError: Either ReadTimeout or server status error
        :raises APIError: error response from petscan
        """
        url = 'https://petscan.wmflabs.org'

        try:
            req = http.fetch(url, params=self.opts)
        except ReadTimeout:
            raise ServerError(f'received ReadTimeout from {url}')

        server_err = HTTPStatus.INTERNAL_SERVER_ERROR
        if server_err <= req.status_code < server_err + 100:
            raise ServerError(
                f'received {req.status_code} status from {req.url}')

        data = req.json()
        if 'error' in data:
            raise APIError('Petscan', data['error'], **self.opts)

        raw_pages = data['*'][0]['a']['*']
        yield from raw_pages

    @property
    def generator(self) -> Generator[pywikibot.page.Page, None, None]:
        """Yield results from :meth:`query`.

        .. versionchanged:: 7.6
           changed from iterator method to generator property
        """
        for raw_page in self.query():
            yield pywikibot.Page(self.site, raw_page['title'],
                                 int(raw_page['namespace']))


class PagePilePageGenerator(GeneratorWrapper):

    """Queries PagePile to generate pages.

    .. seealso:: https://pagepile.toolforge.org/
    .. versionadded:: 9.0
    """

    def __init__(self, id: int) -> None:
        """Initializer.

        :param id: The PagePile id to query
        """
        self.opts = self.buildQuery(id)

    def buildQuery(self, id: int):
        """Get the querystring options to query PagePile.

        :param id: int
        :return: Dictionary of querystring parameters to use in the
            query
        """
        return {
            'id': id,
            'action': 'get_data',
            'format': 'json',
            'doit': ''
        }

    def query(self) -> Generator[str, None, None]:
        """Query PagePile.

        :raises ServerError: Either ReadTimeout or server status error
        :raises APIError: error response from petscan
        """
        url = 'https://pagepile.toolforge.org/api.php'

        req = http.fetch(url, params=self.opts)

        data = req.json()
        if 'error' in data:
            raise APIError('PagePile', data['error'], **self.opts)

        self.site = pywikibot.site.APISite.fromDBName(data['wiki'])
        raw_pages = data['pages']
        yield from raw_pages

    @property
    def generator(self) -> Generator[pywikibot.page.Page, None, None]:
        """Yield results from :meth:`query`."""
        for raw_page in self.query():
            page = pywikibot.Page(self.site, raw_page)
            yield page



"""Family module for Wikimedia chapter, thematic organisation and WUG wikis."""
#
# (C) Pywikibot team, 2012-2024
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from pywikibot import family


class Family(family.SubdomainFamily, family.WikimediaFamily):

    """Family class for WCH, WTO and WUG wikis hosted on wikimedia.org."""

    name = 'wikimediachapter'
    code_aliases = {
        'et': 'ee'
    }

    closed_wikis = ['cn', 'nz', 'pa-us']

    codes = {
        'ae', 'am', 'ar', 'az', 'bd', 'be', 'br', 'ca', 'co', 'dk', 'ec', 'ee',
        'fi', 'ge', 'gr', 'hi', 'id', 'id-internal', 'il', 'mai', 'mk', 'mx',
        'ng', 'nl', 'no', 'nyc', 'pl', 'pt', 'punjabi', 'romd', 'rs', 'ru',
        'se', 'tr', 'ua', 'uk', 've', 'wb',
    }
